# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

##############################################################################

class Library (object):

    CURSECDN_URL_BASE = 'https://modloaders.cursecdn.com/647622546/maven/'

    def __init__(self, library):
        super(Library, self).__init__()
        self.library = library
        parts = self.library['name'].split(':')
        self.group, self.artifact, self.version = parts[:3]
        self.type = parts[3] if len(parts) >= 4 else None
        self.extension = 'jar'
        if '@' in self.version:
            self.version, self.extension = self.version.split('@', 1)

    def get_native_string(self):
        if 'natives' in self.library:
            return self.library['natives']['linux']
        return None

    def get_filename(self):
        nativestring = self.get_native_string()
        if nativestring:
            return '{}-{}-{}.{}'.format(self.artifact, self.version, nativestring, self.extension)
        elif self.type is not None:
            return '{}-{}-{}.{}'.format(self.artifact, self.version, self.type, self.extension)
        return '{}-{}.{}'.format(self.artifact, self.version, self.extension)

    def get_path(self):
        if 'path' in self.library:
            return self.library['path']

        if 'downloads' in self.library:
            nativestring = self.get_native_string()
            if nativestring:
                return self.library['downloads']['classifiers'][nativestring]['path']
            else:
                return self.library['downloads']['artifact']['path']

        return '{}/{}/{}/{}'.format(self.group.replace('.', '/'),
                                    self.artifact,
                                    self.version,
                                    self.get_filename())

    def get_download_url(self):
        if 'downloads' in self.library:
            nativestring = self.get_native_string()
            if nativestring:
                return self.library['downloads']['classifiers'][nativestring]['url']
            else:
                return self.library['downloads']['artifact']['url']
        if 'url' in self.library:
            return '{}{}'.format(self.library['url'], self.get_path())
        return None

    def get_alternate_download_url(self):
        if not self.get_native_string():
            return '{}{}'.format(self.CURSECDN_URL_BASE, self.get_path())
        return None

    def get_artifact(self):
        return '{}:{}'.format(self.group, self.artifact)

    def matches(self, environment):
        if 'rules' not in self.library:
            return True

        allow = False
        for rule in self.library['rules']:
            if self.matches_rule(rule, environment):
                allow = rule['action'] == 'allow'
        return allow

    def matches_rule(self, rule, environment):
        if 'os' in rule:
            return rule['os']['name'] == 'linux'
        return True

    def must_extract(self):
        return 'natives' in self.library

    def extract_excludes(self):
        excludes = []
        if 'extract' in self.library and 'exclude' in self.library['extract']:
            excludes = self.library['extract']['exclude']
        return excludes

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..builder.base import BaseBuilder
from ..framew.cmdapplication import Subcommand
from ..framew.outputter import ListOutputter

##############################################################################


class ModsInfoCommand (Subcommand):
    """
    Display information about the mods in the modpack.
    """

    name = 'modsinfo'

    def setup(self):
        super(ModsInfoCommand, self).setup()
        self.builder = BaseBuilder(self.config)
        self.outputter = ListOutputter()

    def get_cmdline_parser(self):
        parser = super(ModsInfoCommand, self).get_cmdline_parser()
        self.builder.add_cmdline_args(parser)
        self.outputter.add_argument_group(parser)
        return parser

    def run_command(self, arguments):
        self.builder.setup_build(arguments)

        keys = ['Name', 'Filename']
        vals = []

        for mod in self.builder.packlock.get_all_mods():
            vals.append((mod.name, mod.fileName))

        self.outputter.produce_output(arguments, keys, vals)



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import yaml

from ..curse.curseforge import Curseforge
from ..framew.cmdapplication import Subcommand
from ..framew.config import ConfigError
from ..framew.log import getlog

##############################################################################


class ConvertCommand (Subcommand):
    """
    Convert a curse/twitch modpack's manifest to a packmaker yaml
    """

    name = 'convert'

    def setup(self):
        super(ConvertCommand, self).setup()
        self.setup_api()
        self.setup_db()

    def setup_api(self):
        authn_token = self.config.get('curseforge::authentication_token')
        if authn_token is None:
            raise ConfigError('No curseforge authentication token')
        self.api = Curseforge(authn_token)

    def setup_db(self):
        db_filename = self.config.get('curseforge::moddb_filename')
        if db_filename is None:
            raise ConfigError('No moddb_filename parameter in configuration')
        self.db_filename = db_filename

    def get_cmdline_parser(self):
        parser = super(ConvertCommand, self).get_cmdline_parser()
        parser.add_argument('manifest', help='manifest json file')
        parser.add_argument('packdef', nargs='?', default='packmaker.yml', help='output packdef file')
        return parser

    def run_command(self, arguments):
        log = getlog()
        log.info('Reading curseforge manifest file ...')
        with open(arguments.manifest, 'r') as mf:
            manifest = json.load(mf)

        packdef = {'name': manifest['name'],
                   'title': manifest['name'],
                   'version': manifest['version'],
                   'authors': [manifest['author']],
                   'minecraft': manifest['minecraft']['version'],
                   'forge': manifest['minecraft']['modLoaders'][0]['id'].split('-', 1)[1],
                   'mods': [],
                   'resourcepacks': [],
                   'files': []}

        log.info('Looking up addons ...')
        for addon in manifest['files']:
            slug = None
            project_id = addon['projectID']
            file_id = addon['fileID']

            addon = self.api.get_addon(project_id)
            print('DEBUG: addon = {}'.format(addon))
            slug = addon['slug']
            if slug is None:
                print('  !! Cannot find a project with id {}'.format(project_id))
                continue
            addonfile = self.api.get_addon_file(project_id, file_id)
            if addonfile is None:
                print('  || Cannot find the correct version of addon: {} (fileID = {})'.format(slug, file_id))
                continue

            categoryId = int(addon['categorySection']['gameCategoryId'])
            if categoryId == 6:
                packdef['mods'].append({slug: {'version': addonfile['fileName']}})
            elif categoryId == 12:
                packdef['resourcepacks'].append({slug: {'version': addonfile['fileName']}})

        if 'overrides' in manifest:
            packdef['files'].append({'location': manifest['overrides']})

        log.info('Writing out packdef file: {} ...'.format(arguments.packdef))
        with open(arguments.packdef, 'w') as pf:
            yaml.safe_dump(packdef, stream=pf, indent=2, default_flow_style=False)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import requests

from .framew.application import OperationError
from .libraries import Library

##############################################################################


class Minecraft (object):

    VERSION_MANIFEST_URL = 'https://launchermeta.mojang.com/mc/game/version_manifest.json'

    ASSETS_BASE_URL = 'http://resources.download.minecraft.net/'

    def __init__(self, minecraft_version):
        super(Minecraft, self).__init__()
        self.minecraft_version = minecraft_version
        self.manifest = None
        self.assetsindex = None

    def get_version_manifest(self):
        if self.manifest is None:
            self.download_version_manifest()
        return self.manifest

    def get_assets_index(self):
        if self.assetsindex is None:
            self.download_assets_index()
        return self.assetsindex

    def download_client(self, downloader):
        print('Downloading minecraft client ...')
        self.download_client_jar(downloader)
        print('Downloading minecraft client libraries ...')
        self.download_libraries(downloader)
        print('Downloading minecraft client assets ...')
        self.download_assets(downloader)

    def download_server(self, downloader):
        print('Downloading minecraft server ...')
        self.download_server_jar(downloader)
        print('Downloading minecraft server libraries ...')
        self.download_libraries(downloader)

    def get_libraries(self):
        m = self.get_version_manifest()
        libs = []
        for lib in m['libraries']:
            library = Library(lib)
            if library.matches(None):
                libs.append(library)
        return libs

    def get_launch_arguments(self):
        m = self.get_version_manifest()
        if 'arguments' in m:
            return ' '.join([arg for arg in m['arguments']['game'] if type(arg) == str])
        elif 'minecraftArguments' in m:
            return m['minecraftArguments']

    def get_mainclass(self):
        version_info = self.get_version_manifest()
        return version_info['mainClass']

    ##########################################################################

    def download_version_manifest(self):
        manifesturl = None
        meta = requests.get(self.VERSION_MANIFEST_URL).json()
        for version in meta['versions']:
            if version['id'] == self.minecraft_version:
                manifesturl = version['url']
                break
        if manifesturl is None:
            raise OperationError('Cannot find a minecraft manifest for version {}'.format(self.minecraft_version))

        self.manifest = requests.get(manifesturl).json()

    def download_assets_index(self):
        m = self.get_version_manifest()
        assetsurl = m['assetIndex']['url']
        self.assetsindex = requests.get(assetsurl).json()

    def download_client_jar(self, downloader):
        m = self.get_version_manifest()
        clienturl = m['downloads']['client']['url']

        localdir = os.path.join('versions', m['id'])
        if not os.path.exists(os.path.join(downloader.download_dir, localdir)):
            os.makedirs(os.path.join(downloader.download_dir, localdir))

        localfile = os.path.join(localdir, 'client.jar')
        return downloader.download(clienturl, localfile)

    def download_server_jar(self, downloader):
        m = self.get_version_manifest()
        serverurl = m['downloads']['server']['url']

        localdir = os.path.join('versions', m['id'])
        if not os.path.exists(os.path.join(downloader.download_dir, localdir)):
            os.makedirs(os.path.join(downloader.download_dir, localdir))

        localfile = os.path.join(localdir, 'server.jar')
        return downloader.download(serverurl, localfile)

    def download_libraries(self, downloader):
        m = self.get_version_manifest()
        for lib in m['libraries']:
            library = Library(lib)
            if library.matches(None):
                url = library.get_download_url()
                if url is not None:
                    path = library.get_path()

                    localdir = os.path.join(downloader.download_dir, 'libraries', os.path.dirname(path))
                    if not os.path.exists(localdir):
                        os.makedirs(localdir)

                    localfile = os.path.join('libraries', path)
                    downloader.download(url, localfile)

    def download_assets(self, downloader):
        assets = self.get_assets_index()

        assetIndexDir = os.path.join(downloader.download_dir, 'assets', 'indexes')
        if not os.path.exists(assetIndexDir):
            os.makedirs(assetIndexDir)

        assetIndexFile = os.path.join(assetIndexDir, '{}.json'.format(self.get_version_manifest()['assets']))
        with open(assetIndexFile, 'w') as aif:
            aif.write(json.dumps(assets))

        for key, val in assets['objects'].items():
            h = val['hash']
            path = '{}/{}'.format(h[0:2], h)

            localfile = os.path.join('assets', 'objects', path)
            if not os.path.exists(os.path.join(downloader.download_dir, os.path.dirname(localfile))):
                os.makedirs(os.path.join(downloader.download_dir, os.path.dirname(localfile)))

            url = self.ASSETS_BASE_URL + path

            downloader.download(url, localfile)

##############################################################################


def version_as_integer(version_string):
    major, minor, micro = version_string.split('.')
    return (int(major) * 10000) + (int(minor) * 100) + int(micro)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['commands']

from .build import BuildCurseforge, BuildLocal, BuildServer, BuildRouthio
from .convert import ConvertCommand
from .findupdates import FindUpdatesCommand
from .info import InfoCommand
from .launch import LaunchCommand
from .lock import LockCommand
from .modsinfo import ModsInfoCommand
from .resourcepacksinfo import ResourcepacksInfoCommand
from .search import SearchCommand
from .updatedb import UpdateDbCommand

##############################################################################

commands = [
             BuildCurseforge,
             BuildLocal,
             BuildRouthio,
             BuildServer,
             ConvertCommand,
             FindUpdatesCommand,
             InfoCommand,
             LaunchCommand,
             LockCommand,
             ModsInfoCommand,
             ResourcepacksInfoCommand,
             SearchCommand,
             UpdateDbCommand,
           ]

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..curse.curseforge import Curseforge
from ..framew.cmdapplication import Subcommand
from ..framew.config import ConfigError
from ..framew.outputter import ListOutputter

##############################################################################


class SearchCommand (Subcommand):
    """
    Search curseforge for mods.
    """

    name = 'search'

    valid_search_categories_ids = [6,   # mods
                                   12]  # resourcepacks

    def setup(self):
        super(SearchCommand, self).setup()
        self.setup_api()
        self.outputter = ListOutputter()

    def setup_api(self):
        authn_token = self.config.get('curseforge::authentication_token')
        if authn_token is None:
            raise ConfigError('No curseforge authentication token')
        self.api = Curseforge(authn_token)

    def get_cmdline_parser(self):
        parser = super(SearchCommand, self).get_cmdline_parser()
        parser.add_argument('searchstring', nargs='+', help='search string')
        self.outputter.add_argument_group(parser)
        return parser

    def run_command(self, arguments):
        results = ((mod['name'], mod['slug'], mod['categorySection']['name'], mod['id'])
                   for mod in self.api.yield_addons_by_criteria(gameId=432,
                                                                searchFilter=' '.join(arguments.searchstring))
                   if mod['categorySection']['gameCategoryId'] in self.valid_search_categories_ids)

        self.outputter.produce_output(arguments, ('name', 'slug', 'category', 'id'), results)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Base object class for many classes. Provides standard class configuration
utility methods often needed.
"""

__all__ = ["BaseObject"]

##############################################################################


class NonStdlibError (Exception):
    """Base for Exceptions NOT from the python stdlib"""
    pass


class BaseObjectError (NonStdlibError):
    """Top level exception for errors to do with the BaseObject"""
    pass


class InvalidOptions (BaseObjectError):
    """Invalid option definitions in the class"""
    pass


class InvalidParameter (BaseObjectError):
    """Invalid option passed to the constructor of the class"""
    pass

##############################################################################


class BaseObject (object):

    options = {}

    def __init__(self, **kw):
        super(BaseObject, self).__init__()
        self._parse_options(BaseObject.options, kw)

    def _parse_options(self, options, kw):
        for key, v in options.items():
            conversion = None
            if type(v) in (list, tuple):
                try:
                    default_value = v[0]
                    conversion = v[1]
                except LookupError:
                    raise InvalidOptions("Invalid option: %s.%s" % (self.__class__, key))
            else:
                default_value = v

            value = kw.get(key, default_value)
            if conversion is not None:
                try:
                    value = conversion(value)
                except Exception as e:
                    raise InvalidParameter("The parameter '%s.%s=%s' cannot be type converted: %s" % (self.__class__,
                                                                                                      key,
                                                                                                      value,
                                                                                                      e))

            self.__dict__[key] = value

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os.path
import requests
import shutil
import subprocess
import zipfile

from .framew.application import OperationError
from .framew.log import getlog
from .libraries import Library
from .minecraft import Minecraft

##############################################################################


class Forge (object):

    FORGE_REPO_URL = 'https://files.minecraftforge.net/maven/'

    INCOMPATIBLE_FORGE_VERSIONS = ['14.23.5.2851']

    BUILD_THAT_CHANGED_UNIVERSAL_FILENAME = 2851

    ##########################################################################

    def __init__(self, minecraft_version, forge_version, downloader):
        super(Forge, self).__init__()
        self.minecraft_version = minecraft_version
        self.forge_version = forge_version
        self.full_version = '{}-{}'.format(minecraft_version, forge_version)
        self.downloader = downloader
        self.installer_jar_filename = None
        self.install_profile = None
        self.version_info = None

        if self.forge_version in Forge.INCOMPATIBLE_FORGE_VERSIONS:
            raise OperationError('Sorry. Forge version {} does not work with '.format(self.forge_version)
                                 + 'packmaker. Upgrade to a newer version if possible.')

        self.log = getlog()

    ##########################################################################

    def installer_filename(self):
        return 'forge-{}-installer.jar'.format(self.full_version)

    ##########################################################################

    def installer_path(self):
        return 'net/minecraftforge/forge/{}/{}'.format(self.full_version,
                                                       self.installer_filename())

    ##########################################################################

    def installer_download_url(self):
        return '{}/{}'.format(Forge.FORGE_REPO_URL, self.installer_path())

    ##########################################################################

    def universal_filename(self):
        if self.minecraft_version == '1.12.2':
            build_num = int(self.forge_version.split('.')[-1])
            if build_num >= Forge.BUILD_THAT_CHANGED_UNIVERSAL_FILENAME:
                return 'forge-{}.jar'.format(self.full_version)
        return 'forge-{}-universal.jar'.format(self.full_version)

    ##########################################################################

    def universal_path(self):
        return 'net/minecraftforge/forge/{}/{}'.format(self.full_version,
                                                       self.universal_filename())

    ##########################################################################

    def universal_download_url(self):
        return '{}/{}'.format(Forge.FORGE_REPO_URL, self.universal_path())

    ##########################################################################

    def get_install_profile(self):
        if self.install_profile is None:
            installer_filename = self.get_installer_jar()
            with zipfile.ZipFile(installer_filename, 'r') as installerjar:
                with installerjar.open('install_profile.json', 'r') as installf:
                    self.install_profile = json.load(installf)
        return self.install_profile

    ##########################################################################

    def get_installer_jar(self):
        if self.installer_jar_filename is None:
            self.installer_jar_filename = self.downloader.download(self.installer_download_url())
        return self.installer_jar_filename

    ##########################################################################

    def get_universal_jar(self, filename=None):
        installer_filename = self.get_installer_jar()
        if filename is None:
            filename = os.path.join(os.path.dirname(installer_filename), self.universal_filename())

        with zipfile.ZipFile(installer_filename, 'r') as installerjar:
            try:
                universal_filename = installerjar.getinfo(self.universal_filename())
            except KeyError:
                try:
                    universal_filename = installerjar.getinfo('maven/{}'.format(self.universal_path()))
                except KeyError:
                    raise Exception('Cannot find {} in the forge jar file: {}'
                                    .format(self.universal_filename(), installer_filename))

            with open(filename, 'wb') as outjar:
                with installerjar.open(universal_filename, 'r') as injar:
                    shutil.copyfileobj(injar, outjar)

    ##########################################################################

    def get_version_info(self):
        if self.version_info is not None:
            return self.version_info

        installer_filename = self.get_installer_jar()
        with zipfile.ZipFile(installer_filename, 'r') as installerjar:
            try:
                with installerjar.open('version.json', 'r') as installf:
                    self.version_info = json.load(installf)
            except KeyError:
                install_profile = self.get_install_profile()
                if 'versionInfo' in install_profile:
                    self.version_info = install_profile['versionInfo']

        return self.version_info

    ##########################################################################

    def get_launch_arguments(self):
        version_info = self.get_version_info()
        if 'arguments' in version_info:
            return ' '.join([arg for arg in version_info['arguments']['game'] if type(arg) == str])

        elif 'minecraftArguments' in version_info:
            return version_info['minecraftArguments']

    ##########################################################################

    def get_mainclass(self):
        version_info = self.get_version_info()
        return version_info['mainClass']

    ##########################################################################

    def get_dependent_libraries(self):
        dependents = []
        version_info = self.get_version_info()
        libraries = version_info['libraries']
        for lib in libraries:
            library = Library(lib)
            if library.matches(None):
                dependents.append(library)
        return dependents

    def get_installer_libraries(self):
        dependents = []
        install_profile = self.get_install_profile()
        if 'libraries' in install_profile:
            for lib in install_profile['libraries']:
                library = Library(lib)
                if library.matches(None):
                    dependents.append(library)
        return dependents

    ##########################################################################

    def install_dependent_libraries(self, cache_location, install_location):
        dependent_libs = self.get_dependent_libraries()
        self.install_libraries(cache_location, install_location, dependent_libs)

    def install_installer_libraries(self, cache_location, install_location):
        installer_libs = self.get_installer_libraries()
        self.install_libraries(cache_location, install_location, installer_libs)

    def install_libraries(self, cache_location, install_location, libraries):
        for library in libraries:
            cached_library = os.path.join(self.downloader.download_dir, library.get_path())

            url = library.get_alternate_download_url()
            try:
                self.downloader.download(url, library.get_path())
            except requests.exceptions.HTTPError:
                url = library.get_download_url()
                if url is not None:
                    self.downloader.download(url, library.get_path())
                else:
                    continue

            local_library = os.path.join(install_location, library.get_path())
            if not os.path.exists(os.path.dirname(local_library)):
                os.makedirs(os.path.dirname(local_library))

            shutil.copy2(cached_library, local_library)

    def run_processors(self, install_location, side):
        install_profile = self.get_install_profile()

        if 'processors' not in install_profile:
            return

        self.log.info('Processing installer data ...')
        data = {}
        if 'data' in install_profile:
            for datakey, dataval in install_profile['data'].items():
                dataval = dataval[side]
                if dataval.startswith('[') and dataval.endswith(']'):
                    # artifact
                    data[datakey] = os.path.join(install_location,
                                                 mavenname_to_path(dataval[1:-1]))

                elif dataval.startswith('\'') and dataval.endswith('\''):
                    # literal
                    data[datakey] = dataval[1:-1]

                else:
                    self.log.info('  Extracting: {}'.format(dataval))

                    data_location = os.path.join(install_location, 'data')
                    if not os.path.exists(data_location):
                        os.makedirs(data_location)

                    installerjar = self.get_installer_jar()
                    with zipfile.ZipFile(installerjar, 'r') as installer:
                        installer.extract(dataval.lstrip('/'), data_location)

                    data[datakey] = os.path.join(data_location, dataval.lstrip('/'))

        manifest = Minecraft(self.minecraft_version).get_version_manifest()

        data['SIDE'] = side
        data['MINECRAFT_JAR'] = os.path.join(install_location, 'com', 'mojang', 'minecraft',
                                             manifest['id'], 'minecraft-{}-{}.jar'.format(manifest['id'], side))

        self.log.info('Executing forge installer processors ...')
        processors = install_profile['processors']

        i = 0
        for proc in processors:
            i += 1
            self.log.info('  Running {} of {} processors ...'.format(i, len(processors)))

            classpath = proc['classpath']
            args = proc['args']

            finalargs = []
            for arg in args:
                if arg.startswith('{') and arg.endswith('}'):
                    arg = data[arg[1:-1]]
                elif arg.startswith('[') and arg.endswith(']'):
                    arg = os.path.join(install_location,
                                       mavenname_to_path(arg[1:-1]))

                finalargs.append(arg)

            classpath = [os.path.join(install_location, mavenname_to_path(proc['jar']))]
            for lib in proc['classpath']:
                classpath.append(os.path.join(install_location, mavenname_to_path(lib)))

            mainclass = find_mainclass_in_jar(os.path.join(install_location,
                                                           mavenname_to_path(proc['jar'])))

            cmd = ['java', '-cp', ':'.join(classpath), mainclass]
            cmd.extend(finalargs)

            print('DEBUG: cmd = {}'.format(' '.join(cmd)))
            result = subprocess.run(cmd)
            if result.returncode != 0:
                raise OperationError('Processor finished with errors. Abort')

##############################################################################


def mavenname_to_path(name):
    parts = name.split(':')
    filename = '{}-{}'.format(parts[2], parts[3]) if len(parts) >= 4 else parts[2]
    filename = filename.replace('@', '.') if '@' in filename else '{}.jar'.format(filename)
    initpath = parts[0].split('.')
    initpath.append(parts[1])
    initpath.append(parts[2].split('@')[0])
    initpath.append('{}-{}'.format(parts[1], filename))
    return '/'.join(initpath)


def find_mainclass_in_jar(jarfile):
    with zipfile.ZipFile(jarfile, 'r') as jar:
        with jar.open('META-INF/MANIFEST.MF', 'r') as manifest:
            lines = manifest.readlines()
            print('DEBUG: lines = {}'.format(lines))
            for line in lines:
                line = line.decode('utf-8').strip()
                if line.startswith('Main-Class:'):
                    return line.split(' ')[1]
    return None

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os.path
import re

from .framew.application import OperationError
from .framew.log import getlog

##############################################################################


class PackLock (object):

    SUBST_REX = re.compile(r'\${(\w+)(:-(.+))?}')

    def __init__(self, filenames):
        super(PackLock, self).__init__()
        self.log = getlog()
        if not isinstance(filenames, (list, tuple)):
            filenames = [filenames]
        self.lock_filenames = filenames
        self.lockedmods = {}
        self.lockedreps = {}
        self.metadata = {}
        self.files = []
        self.files_fullpath = []
        self.extraopt = {}

    def load(self):
        for filename in self.lock_filenames:
            self.load_one(filename)

    def load_one(self, filename):
        self.log.info('Loading packlock file: {}'.format(filename))
        with open(filename, 'r') as pl:
            try:
                lockdict = json.load(pl)
            except json.decoder.JSONDecodeError as err:
                raise OperationError('Invalid pack lock file: {}'.format(err))
        self.parse(lockdict, filename)

    def parse(self, lockdict, filename):
        version = lockdict.get('version', None)
        if version is None:
            self.parse_ver0(lockdict, filename)
        elif version == '1':
            self.parse_ver1(lockdict, filename)
        else:
            raise OperationError('Unknown format for pack lock file: version = {}'.format(version))

    def parse_ver0(self, lockdict, filename):
        # pre-versioned format. Only has locked mods in it
        for slug in lockdict.keys():
            resolveddict = lockdict[slug]
            self.lockedmods[slug] = LockedElement(slug, resolveddict, 'mod')

    def parse_ver1(self, lockdict, filename):
        metadata = lockdict.get('metadata', {})
        if metadata:
            for m in ('name', 'title', 'version', 'authors', 'minecraft_version', 'forge_version'):
                if m in metadata and metadata[m]:
                    val = self.parse_environ_variables(metadata[m])
                    self.set_metadata(m, val)

        parsed_files = lockdict.get('files', [])
        if parsed_files:
            self.files.extend(parsed_files)
            lock_basedir = os.path.dirname(os.path.abspath(filename))
            for parsed_file in parsed_files:
                parsed_location = parsed_file['location']
                if not parsed_location.startswith(os.path.sep):
                    parsed_location = os.path.join(lock_basedir, parsed_location)
                self.files_fullpath.append({'location': parsed_location,
                                            'clientonly': parsed_file.get('clientonly', False),
                                            'serveronly': parsed_file.get('serveronly', False)
                                            })

        self.extraopt.update(lockdict.get('extraopt', {}))
        resolutions = lockdict.get('resolutions')

        mod_resolutions = {}
        rep_resolutions = {}
        for slug, resolution in resolutions.items():
            if 'type' not in resolution:
                mod_resolutions[slug] = resolution
            else:
                if resolution['type'] == 'mod':
                    mod_resolutions[slug] = resolution
                elif resolution['type'] == 'resourcepack':
                    rep_resolutions[slug] = resolution
                else:
                    raise Exception('Unknown resolution type for "{}": {}'.format(slug, resolution['type']))

        self.parse_ver0(mod_resolutions, filename)

        for slug, resolveddict in rep_resolutions.items():
            self.lockedreps[slug] = LockedElement(slug, resolveddict, 'resourcepack')

    def parse_environ_variables(self, val):
        if isinstance(val, list):
            for n in range(len(val)):
                val[n] = self.parse_environ_variables(val[n])

        elif isinstance(val, dict):
            for k, v in val.items():
                val[k] = self.parse_environ_variables(val[k])

        else:
            subst = PackLock.SUBST_REX.search(val)
            while subst:
                var = subst.group(1)
                try:
                    substval = os.environ[var]
                except KeyError:
                    substval = subst.group(3)
                    if substval is None:
                        raise OperationError('Undefined environment variable in pack lock file: {}'.format(var))
                val = re.sub(r'\${%s}' % subst.group(0)[2:-1], substval, val)
                subst = PackLock.SUBST_REX.search(val)

        return val

    def save(self):
        resolutions = {}
        for slug, mod in self.lockedmods.items():
            resolutions[slug] = mod.as_dict()
        for slug, rep in self.lockedreps.items():
            resolutions[slug] = rep.as_dict()

        lockdata = {'version': '1',
                    'metadata': self.metadata,
                    'resolutions': resolutions,
                    'files': self.files,
                    'extraopt': self.extraopt
                    }
        lock = json.dumps(lockdata)
        with open(self.lock_filenames[0], 'w') as lockfile:
            lockfile.write(lock)

    def add_resolved_mod(self, moddef, modobj):
        slug = moddef.slug
        if isinstance(modobj, LockedElement):
            self.lockedmods[slug] = modobj
        else:
            self.lockedmods[slug] = LockedElement(slug, modobj, 'mod', moddef)

    def add_resolved_resourcepack(self, repdef, repobj):
        slug = repdef.slug
        if isinstance(repobj, LockedElement):
            self.lockedreps[slug] = repobj
        else:
            self.lockedreps[slug] = LockedElement(slug, repobj, 'resourcepack', repdef)

    def add_files(self, filesdef):
        self.files.append(filesdef)

    def add_extraopt(self, key, extraopt):
        self.extraopt[key] = extraopt

    def set_all_metadata(self, packdef):
        if packdef.name:
            self.set_metadata('name', packdef.name)
        if packdef.title:
            self.set_metadata('title', packdef.title)
        if packdef.version:
            self.set_metadata('version', packdef.version)
        if packdef.authors:
            self.set_metadata('authors', packdef.authors)
        if packdef.minecraft_version:
            self.set_metadata('minecraft_version', packdef.minecraft_version)
        if packdef.forge_version:
            self.set_metadata('forge_version', packdef.forge_version)

    def set_metadata(self, key, value):
        self.metadata[key] = value

    def get_mod(self, slug):
        return self.lockedmods[slug]

    def get_all_mods(self):
        return self.lockedmods.values()

    def get_resourcepack(self, slug):
        return self.lockedreps[slug]

    def get_all_resourcepacks(self):
        return self.lockedreps.values()

    def yield_clientonly_mods(self, include_optional=False):
        for mod in self.lockedmods.values():
            if not include_optional and mod.optional is True:
                continue
            if mod.serveronly is True:
                continue
            yield mod

    def yield_serveronly_mods(self, include_optional=False):
        for mod in self.lockedmods.values():
            if not include_optional and mod.optional is True:
                continue
            if mod.clientonly is True:
                continue
            yield mod

    def yield_clientonly_files(self):
        for files in self.files_fullpath:
            if files['serveronly'] is True:
                continue
            yield files

    def yield_serveronly_files(self):
        for files in self.files_fullpath:
            if files['clientonly'] is True:
                continue
            yield files

    def get_extraopt(self, key):
        return self.extraopt.get(key, None)

    def get_metadata(self, key):
        return self.metadata.get(key, None)

    def get_all_metadata(self):
        return self.metadata

##############################################################################


class LockedElement (object):

    _attrs = ['projectId', 'name', 'fileId', 'fileName', 'author', 'downloadUrl',
              'website', 'version', 'clientonly', 'serveronly', 'optional',
              'recommendation', 'selected', 'description', 'type'
              ]

    def __init__(self, slug, resolveddict, resolvedtype, moddef=None):
        super(LockedElement, self).__init__()
        self.slug = slug
        for attr in self._attrs:
            val = resolveddict.get(attr, None)
            if val is None and moddef is not None and hasattr(moddef, attr):
                val = moddef.__dict__[attr]
            self.__dict__[attr] = val
        self.__dict__['type'] = resolvedtype

    def as_dict(self):
        asdict = {}
        for attr in self._attrs:
            asdict[attr] = self.__dict__[attr]
        return asdict

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import requests
import urllib.parse

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from .framew.application import OperationError
from .framew.log import getlog

log = getlog()

##############################################################################


def requests_retry_session(retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504), session=None):
    session = session or requests.Session()
    retry = Retry(total=retries, read=retries, connect=retries,
                  backoff_factor=backoff_factor,
                  status_forcelist=status_forcelist)
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session

##############################################################################


class HttpDownloader (object):

    def __init__(self, download_dir):
        super(HttpDownloader, self).__init__()
        self.download_dir = download_dir

    ##########################################################################

    def subdownloader(self, download_dir):
        return HttpDownloader(os.path.join(self.download_dir, download_dir))

    ##########################################################################

    def download(self, url, filename=None):
        if filename is None:
            filename = os.path.basename(urllib.parse.urlparse(url).path)
        localfile = os.path.join(self.download_dir, filename)
        if not os.path.exists(localfile):
            log.info('  downloading : {}'.format(url))
            self.do_file_download(url, localfile)
        else:
            log.info('  [cached]    : {}'.format(url))
        return localfile

    ##########################################################################

    def do_file_download(self, url, dest):
        destdir = os.path.dirname(dest)
        if not os.path.exists(destdir):
            os.makedirs(destdir)
        tempfile = os.path.join(destdir, '.downloading_{}'.format(os.path.basename(dest)))
        if os.path.exists(tempfile):
            os.remove(tempfile)
        try:
            with requests_retry_session().get(url, stream=True, timeout=(5, 15)) as r:
                r.raise_for_status()
                with open(tempfile, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=65536):
                        if chunk:
                            f.write(chunk)
            os.rename(tempfile, dest)
        finally:
            try:
                os.remove(tempfile)
            except FileNotFoundError:
                pass

##############################################################################


class MavenDownloader (HttpDownloader):

    def __init__(self, download_dir, maven_repos):
        super(MavenDownloader, self).__init__(download_dir)
        self.maven_repos = maven_repos

    ##########################################################################

    def find_maven_repo(self, library):
        group, artifact, version = library.split(':')
        library_filename = '{}-{}.jar'.format(artifact, version)

        cached_library = os.path.join(self.download_dir, library_filename)
        cached_library_meta = '{}.meta'.format(cached_library)

        if os.path.exists(cached_library_meta):
            with open(cached_library_meta, 'r') as lmf:
                meta = json.load(lmf)
            mavenrepo = meta.get('mavenRepo', None)
            if mavenrepo is not None:
                log.info('  [cached]    : {} - {}'.format(library, mavenrepo))
                return mavenrepo

        localfile, mavenrepo = self.download_library_from_repo(group, artifact, version, library_filename)
        if mavenrepo is None:
            raise OperationError('Cannot find a maven repository for  {}'.format(library))

        meta = {'mavenRepo': mavenrepo}
        with open(cached_library_meta, 'w') as mf:
            mf.write(json.dumps(meta))

        return mavenrepo

    ##########################################################################

    def download_library_from_repo(self, group, artifact, version, filename):
        urlpath = '{}/{}/{}/{}'.format(group.replace('.', '/'), artifact, version, filename)
        localfile = os.path.join(self.download_dir, filename)

        for mavenrepo in self.maven_repos:
            url = '{}{}'.format(mavenrepo, urlpath)
            log.info('  downloading : {}'.format(url))
            try:
                self.do_file_download(url, localfile)
            except requests.exceptions.HTTPError:
                continue
            return localfile, mavenrepo

        return None, None

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests

from ..framew.baseobject import BaseObject

##############################################################################


class CurseAPI (BaseObject):

    max_pagesize = 1000
    max_index = 10000

    def __init__(self):
        super(BaseObject, self).__init__()

    def _api_baseurl(self):
        raise NotImplementedError

    def _request_headers(self):
        return None

    def _getrequest(self, url, params=None):
        r = requests.get(self._api_baseurl() + url, params, timeout=60, headers=self._request_headers())
        r.raise_for_status()
        return r

    def _postrequest(self, url, data):
        r = requests.post(self._api_baseurl() + url, json=data, timeout=60, headers=self._request_headers())
        r.raise_for_status()
        return r

    def get_addon(self, projectId):
        return self._getrequest('addon/{}'.format(projectId)).json()

    def get_addon_file(self, projectId, fileId):
        return self._getrequest('addon/{}/file/{}'.format(projectId, fileId)).json()

    def get_addons(self, projectIds):
        return self._postrequest('addon', projectIds).json()

    def get_addon_files(self, projectId):
        return self._getrequest('addon/{}/files'.format(projectId)).json()

    def get_addons_by_criteria(self,
                               gameId,
                               sectionId=None,
                               categoryIds=None,
                               sort='Featured',
                               sortDescending=True,
                               gameVersions=None,
                               index=0,
                               pageSize=1000,
                               searchFilter=None):
        url = 'addon/search'
        parameters = {"gameID": gameId,
                      "sort": sort,
                      "sortDescending": sortDescending,
                      "index": index,
                      "pageSize": pageSize
                      }

        if sectionId is not None:
            parameters['sectionId'] = sectionId
        if categoryIds is not None:
            parameters['categoryId'] = categoryIds
        if gameVersions is not None:
            parameters['gameVersion'] = gameVersions
        if searchFilter is not None:
            parameters['searchFilter'] = searchFilter

        return self._getrequest(url, parameters).json()

    def yield_addons_by_criteria(self,
                                 gameId,
                                 sectionId=None,
                                 categoryIds=None,
                                 sort='Featured',
                                 sortDescending=True,
                                 gameVersions=None,
                                 pageSize=1000,
                                 searchFilter=None):

        if pageSize > self.max_pagesize:
            raise 'curse api limits pagesize to 1000'

        index = 0
        done = False
        while not done:
            results = self.get_addons_by_criteria(gameId, sectionId, categoryIds,
                                                  sort, sortDescending, gameVersions,
                                                  index, pageSize, searchFilter)
            index += pageSize
            if len(results) < pageSize:
                done = True

            if index > self.max_index:
                done = True
            else:
                if index + pageSize > self.max_index:
                    pageSize = self.max_index - index
                    if pageSize < 1:
                        done = True

            for addon in results:
                yield addon

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .api import CurseAPI

##############################################################################


class Curseforge (CurseAPI):

    CURSE_BASEURL = 'https://addons-ecs.forgesvc.net/api/v2/'

    def __init__(self, authn_token):
        super(Curseforge, self).__init__()
        self.authn_token = authn_token

    def _api_baseurl(self):
        return self.CURSE_BASEURL

    def _request_headers(self):
        return {'AuthenticationToken': self.authn_token,
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit 537.6 ' +
                              '(KHTML, like Gecko) Chrome/76.0.3809.132 Safari/537.6'}

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import dateutil.parser

from ..curse.curseforge import Curseforge
from ..builder.base import BaseBuilder
from ..framew.application import OperationError
from ..framew.cmdapplication import Subcommand
from ..framew.config import ConfigError
from ..framew.log import getlog
from ..framew.outputter import ListOutputter

##############################################################################


class FindUpdatesCommand (Subcommand):
    """
    Search curseforge for newer/updated versions of addons.
    """

    name = 'findupdates'

    def setup(self):
        super(FindUpdatesCommand, self).setup()
        self.setup_api()
        self.builder = BaseBuilder(self.config)
        self.outputter = ListOutputter()

    def setup_api(self):
        authn_token = self.config.get('curseforge::authentication_token')
        if authn_token is None:
            raise ConfigError('No curseforge authentication token')
        self.api = Curseforge(authn_token)

    def get_cmdline_parser(self):
        parser = super(FindUpdatesCommand, self).get_cmdline_parser()
        self.builder.add_cmdline_args(parser)
        self.outputter.add_argument_group(parser)
        return parser

    def run_command(self, parsed_args):
        self.builder.setup_build(parsed_args)

        newaddons = []

        getlog().info('Searching for newer versions of addons ...')

        def all_addons():
            for mod in self.builder.packlock.get_all_mods():
                yield mod
            for resourcepack in self.builder.packlock.get_all_resourcepacks():
                yield resourcepack

        for resolvedaddon in all_addons():
            try:
                addonfile_found = None
                latestTimestamp = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
                for addonfile in self.api.get_addon_files(resolvedaddon.projectId):
                    if self.builder.packlock.get_metadata('minecraft_version') not in addonfile['gameVersion']:
                        continue
                    timestamp = dateutil.parser.parse(addonfile['fileDate'])
                    if timestamp > latestTimestamp:
                        addonfile_found = addonfile
                        latestTimestamp = timestamp

                if addonfile_found is None:
                    raise OperationError('Cannot find a addon file for {}'.format(resolvedaddon.name))

                if addonfile_found['id'] != resolvedaddon.fileId:
                    newaddons.append((resolvedaddon.fileName, addonfile_found['fileName']))
            except KeyError:
                newaddons.append(('{} [NEW]'.format(resolvedaddon.name), ''))

        self.outputter.produce_output(parsed_args,
                                      ('current version', 'new version'),
                                      ((m[0], m[1]) for m in newaddons))

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Standard python logging extensions and configuration classes.
"""

__all__ = ['configure', 'getlog']

import errno
import logging
import os
import sys
import time
import types
import traceback

from .baseobject import BaseObject
from .config import Config
from .string_conversion import convert_to_bool

try:
    import codecs
except ImportError:
    codecs = None

##############################################################################

# Logging extensions

TRACE = (logging.NOTSET + logging.DEBUG) / 2
MOREINFO = (logging.DEBUG + logging.INFO) / 2

# Logger class methods to be injected


def trace_method(self, msg='trace', *args, **kwargs):
    if self.manager.disable >= TRACE:
        return
    if TRACE >= self.getEffectiveLevel():
        stack = traceback.extract_stack()[-2]
        msg = '%s [%s:%i] %s' % (stack[2] + '()',             # FUNCNAME
                                 os.path.basename(stack[0]),  # FILENAME
                                 stack[1],                    # LINENO
                                 msg)
        self.log(*(TRACE, msg, args), **kwargs)
        self.log(logging.Logger.TRACE, msg)


def moreinfo_method(self, msg, *args, **kwargs):
    if self.manager.disable >= MOREINFO:
        return
    if MOREINFO >= self.getEffectiveLevel():
        self.log(*(MOREINFO, msg, args), **kwargs)


def repr_method(self):
    return '<logging.Logger instance at 0x%x (name = "%s")>' % (id(self), self.name)

##############################################################################

# Logging configuration class


class LoggingFactory (BaseObject):

    DEFAULT_LEVEL = 'INFO'
    DEFAULT_FORMAT = '%(asctime)s %(levelname)-8s %(message)s'
    DEFAULT_DATEFMT = '%Y-%m-%d %H:%M:%S'

    DEFAULT_CONSOLE_LEVEL = 'WARNING'
    DEFAULT_CONSOLE_FORMAT = '%(message)s'
    DEFAULT_CONSOLE_DATEFMT = ''

    options = {'file':            None,
               'level':           DEFAULT_LEVEL,
               'format':          DEFAULT_FORMAT,
               'datefmt':         DEFAULT_DATEFMT,
               'console':         (True, convert_to_bool),
               'console_level':   DEFAULT_CONSOLE_LEVEL,
               'console_format':  DEFAULT_CONSOLE_FORMAT,
               'console_datefmt': DEFAULT_CONSOLE_DATEFMT,
               'enabled':         (True, convert_to_bool),
               }

    def __init__(self, **kw):
        super(LoggingFactory, self).__init__()
        self._parse_options(LoggingFactory.options, kw)
        self.is_setup = False
        self.setup()
        self.loggers = dict()

    def setup(self):
        logging.getLogger('').handlers = []

        logging.addLevelName(TRACE, 'TRACE')
        logging.addLevelName(MOREINFO, 'MOREINFO')

        if self.enabled is True:

            # setup root logger
            self.root_logger = logging.getLogger('')
            if self.file is not None:
                formatter = logging.Formatter(self.format, self.datefmt)
                handler = MyFileHandler(self.file, 'a')
                handler.setFormatter(formatter)
                self.root_logger.addHandler(handler)
            else:
                self.root_logger.addHandler(DevnullHandler())
            self.root_logger.setLevel(logging.getLevelName(self.level))

            setattr(self.root_logger, 'trace', trace_method)
            setattr(self.root_logger, 'moreinfo', moreinfo_method)
            setattr(self.root_logger, '__repr__', repr_method)

            # setup a console logger, if enabled
            if self.console is True:
                console_fmtr = logging.Formatter(self.console_format, self.console_datefmt)
                console_hdlr = logging.StreamHandler(sys.stderr)
                console_hdlr.setFormatter(console_fmtr)
                console_hdlr.setLevel(logging.getLevelName(self.console_level))
                self.root_logger.addHandler(console_hdlr)

        self.is_setup = True

    def get_logger(self, name=None, config=None):
        if self.enabled is False:
            logger = self.devnull_logger()
        elif name is None:
            logger = self.root_logger
        elif name in self.loggers:
            logger = self.loggers[name]
        else:
            logger = self.make_logger(name, config)
            self.loggers[name] = logger

        return logger

    def make_logger(self, name, config):
        logger = logging.getLogger(name)
        if name is not None:
            logfile = config.get('file', None)
            if logfile is not None:
                handler = MyFileHandler(logfile, 'a')
                format = config.get('format', self.root_format)
                datefmt = config.get('datefmt', self.root_datefmt)
                formatter = logging.Formatter(format, datefmt)
                handler.setFormatter(formatter)
                logger.addHandler(handler)
            level = config.get('level', None)
            if level is not None:
                logger.setLevel(logging.getLevelName(level))

        if not hasattr(logger, 'trace'):
            logger.trace = types.MethodType(trace_method, logger, logger.__class__)
        if not hasattr(logger, 'moreinfo'):
            logger.moreinfo = types.MethodType(moreinfo_method, logger, logger.__class__)
        if not hasattr(logger, '__repr__'):
            logger.__repr__ = types.MethodType(repr_method, logger, logger.__class__)

        return logger

    def devnull_logger(self):
        try:
            return self._devnull_logger
        except AttributeError:
            self._devnull_logger = logging.getLogger("devnull")
            self._devnull_logger.addHandler(DevnullHandler())
            self._devnull_logger.propagate = False
            return self._devnull_logger

##############################################################################

# Logging write handlers


class DevnullHandler (logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)

    def handle(self, record):
        pass


class MyFileHandler (logging.StreamHandler):

    checkperiod = 10  # seconds

    def __init__(self, filename, mode='a', encoding=None):
        if codecs is None:
            encoding = None
        self.baseFilename = os.path.abspath(filename)
        self.mode = mode
        self.encoding = encoding
        logging.StreamHandler.__init__(self, self._open())

    def close(self):
        if self.stream:
            try:
                self.flush()
            except ValueError:
                pass
            if hasattr(self.stream, 'close'):
                self.stream.close()
            logging.StreamHandler.close(self)
            self.stream = None

    def _open(self):
        if self.encoding is None:
            stream = open(self.baseFilename, self.mode)
        else:
            stream = codecs.open(self.baseFilename, self.mode, self.encoding)
        self.inode = os.stat(self.baseFilename).st_ino
        self.lastcheck = time.time()
        return stream

    def _check(self):
        now = time.time()
        if self.lastcheck + self.checkperiod >= now:
            return

        try:
            inode = os.stat(self.baseFilename).st_ino
        except OSError as err:
            if err.errno != errno.ENOENT:
                raise
            # File was probably deleted.
            inode = None

        if inode != self.inode:
            self.close()
            self.stream = self._open()
        self.lastcheck = now

    def emit(self, record):
        self._check()
        try:
            msg = self.format(record)
            if msg.endswith('\n'):
                fs = '%s'
            else:
                fs = '%s\n'
            if not hasattr(types, "UnicodeType"):
                self.stream.write(fs % msg)
            else:
                try:
                    self.stream.write(fs % msg)
                except UnicodeError:
                    self.stream.write(fs % msg.encode("UTF-8"))
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except: # noqa
            self.handleError(record)


##############################################################################

__logging_factory = None


def configure(config, cmdline_arguments):
    global __logging_factory
    if isinstance(config, Config) and 'logging' in config.get_sections():
        logconfig = config.section('logging')
    else:
        logconfig = {}

    if cmdline_arguments.logfile:
        logconfig['file'] = cmdline_arguments.logfile

    console_level = {0: 'WARNING',
                     1: 'INFO',
                     2: 'MOREINFO',
                     3: 'DEBUG',
                     4: 'TRACE',
                     }.get(cmdline_arguments.verbose_level, 'TRACE')
    logconfig['console_level'] = console_level

    __logging_factory = LoggingFactory(**logconfig)


def getlog(name=None, config=None):
    global __logging_factory
    if __logging_factory is None:
        return logging
    return __logging_factory.get_logger(name, config)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

from .base import BaseBuilder
from ..download import HttpDownloader
from ..forge import Forge
from ..minecraft import Minecraft, version_as_integer

##############################################################################


class ServerBuilder (BaseBuilder):

    build_subloc = 'server'

    MAVEN_REPOS = ('https://libraries.minecraft.net/',
                   'https://repo1.maven.org/maven2/',
                   'https://files.minecraftforge.net/maven/'
                   )

    temp_server_script = """
#!/bin/sh

MAX_RAM="4096M"

JAVA_PARAMETERS="-XX:+UseG1GC -Dsun.rmi.dgc.server.gcInterval=2147483646"
JAVA_PARAMETERS="${{JAVA_PARAMETERS}} -XX:+UnlockExperimentalVMOptions"
JAVA_PARAMETERS="${{JAVA_PARAMETERS}} -XX:G1NewSizePercent=20"
JAVA_PARAMETERS="${{JAVA_PARAMETERS}} -XX:G1ReservePercent=20"
JAVA_PARAMETERS="${{JAVA_PARAMETERS}} -XX:MaxGCPauseMillis=50"
JAVA_PARAMETERS="${{JAVA_PARAMETERS}} -XX:G1HeapRegionSize=32M"
JAVA_PARAMETERS="${{JAVA_PARAMETERS}} -Dfml.readTimeout=180"

SERVER_JAR="{}"

echo "Starting minecraft server..."
echo java -server -Xmx${{MAX_RAM}} ${{JAVA_PARAMETERS}} -jar ${{SERVER_JAR}} nogui
exec java -server -Xmx${{MAX_RAM}} ${{JAVA_PARAMETERS}} -jar ${{SERVER_JAR}} nogui
"""

    ##########################################################################

    def setup_build(self, parsed_args):
        super(ServerBuilder, self).setup_build(parsed_args)

        self.mods_location = self.build_location('mods')
        self.libs_location = self.build_location('libraries')

    def do_build(self):
        self.downloader = HttpDownloader(self.cache_location())

        self.log.info('Copying local files ...')
        self.copy_files(files_iterator=self.packlock.yield_serveronly_files())

        self.log.info('Copying mod files ...')
        self.copy_mods(self.mods_location, self.packlock.yield_serveronly_mods())

        mc = Minecraft(self.packlock.get_metadata('minecraft_version'))
        server_jarfile = self.copy_minecraft_server(mc)

        if self.packlock.get_metadata('forge_version') is not None:
            server_jarfile = self.copy_forge(mc)

        self.log.info('Creating stoopid simple launch script (to be replaced soon) ...')
        launch_script = os.path.join(self.build_location(), 'start.sh')
        with open(launch_script, 'w') as startf:
            startf.write(self.temp_server_script.format(server_jarfile))
        os.chmod(launch_script, 0o755)

    ##########################################################################

    def copy_minecraft_server(self, mc):
        mc.download_server(self.downloader.subdownloader('minecraft'))
        manifest = mc.get_version_manifest()

        # Copy the minecraft server library into place.

        cached_lib = os.path.join(self.cache_location('minecraft'), 'versions', manifest['id'], 'server.jar')
        if version_as_integer(self.packlock.get_metadata('minecraft_version')) < 11400:
            local_lib = os.path.join(self.build_location(), 'minecraft_server.{}.jar'.format(manifest['id']))
        else:
            local_lib = os.path.join(self.libs_location, 'com', 'mojang', 'minecraft', manifest['id'],
                                     'minecraft-{}-server.jar'.format(manifest['id']))

        if not os.path.exists(os.path.dirname(local_lib)):
            os.makedirs(os.path.dirname(local_lib))

        shutil.copy2(cached_lib, local_lib)
        server_jarfile = local_lib

        # Copy the minecraft libraries into place.

        for lib in mc.get_libraries():
            cached_lib = os.path.join(self.cache_location('minecraft'), 'libraries', lib.get_path())
            if lib.must_extract():
                continue

            localdir = os.path.join(self.libs_location, os.path.dirname(lib.get_path()))
            if not os.path.exists(localdir):
                os.makedirs(localdir)

            local_lib = os.path.join(self.libs_location, lib.get_path())
            shutil.copy2(cached_lib, local_lib)

        return server_jarfile

    ##########################################################################

    def copy_forge(self, mc):
        mc.get_version_manifest()

        forgedl = Forge(self.packlock.get_metadata('minecraft_version'),
                        self.packlock.get_metadata('forge_version'),
                        self.downloader.subdownloader('forge'))

        self.log.info('Installing forge and dependent libraries ...')
        forgedl.install_dependent_libraries(self.cache_location('library'), self.libs_location)

        installer_location = self.libs_location
        forgedl.install_installer_libraries(self.cache_location('library'), installer_location)

        forgedl.run_processors(installer_location, 'server')

        server_jarfile = 'forge_server-{}.jar'.format(forgedl.full_version)

        if version_as_integer(self.packlock.get_metadata('minecraft_version')) < 11400:
            self.log.info('Extract universal forge jar from installer...')
            universal_jar = os.path.join(self.build_location(), server_jarfile)
            forgedl.get_universal_jar(universal_jar)

        else:
            src_jarfile = os.path.join(self.libs_location,
                                       'net/minecraftforge/forge/{}/forge-{}.jar'.format(forgedl.full_version,
                                                                                         forgedl.full_version))
            dst_jarfile = os.path.join(self.build_location(), server_jarfile)
            shutil.copy2(src_jarfile, dst_jarfile)

        return server_jarfile

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple string conversion utilities.
"""

__all__ = ['ConversionError', 'convert_to_bool', 'convert_to_seconds',
           'convert_to_milliseconds', 'convert_to_minutes', 'convert_to_hours',
           'convert_to_days', 'convert_to_bytes', 'convert_to_kilobytes',
           'convert_to_megabytes', 'convert_to_gigabytes', 'convert_to_terabytes',
           'convert_to_integer', 'convert_to_octal', 'convert_to_floating',
           'convert_to_whitespace_list', 'convert_to_comma_list',
           'convert_to_list', 'convert_to_uid', 'convert_to_gid',
           ]

import re

from .baseobject import NonStdlibError

##############################################################################


class ConversionError(NonStdlibError):
    """Raised when one of these function fails to convert a string"""
    pass


time_string_regex = re.compile(r'([\d.]+)\s*(ms|s|m|h|d)?')
size_string_regex = re.compile(r'([\d.]+)\s*(b|k|m|g|t)?')

##############################################################################


def convert_to_bool(string, allow_invalid_strings=True, invalid_string_value=False):
    if string is None or type(string) is bool:
        return string
    string = string.lower()
    if string in ('1', 'true', 'yes', 'on'):
        return True
    elif string in ('0', 'false', 'no', 'off'):
        return False
    if allow_invalid_strings:
        return invalid_string_value
    else:
        raise ConversionError


def convert_to_milliseconds(string, default_unit='ms'):
    if string is None or type(string) in (float, int):
        return string
    string = string.lower()
    match = time_string_regex.match(string)
    if match is None:
        raise ConversionError
    tvalue = float(match.group(1))
    tunit = match.group(2)
    if tunit is None:
        tunit = default_unit
    if tunit == 'ms':
        return tvalue
    elif tunit == 's':
        return tvalue * 1000
    elif tunit == 'm':
        return tvalue * 60000
    elif tunit == 'h':
        return tvalue * 3600000
    elif tunit == 'd':
        return tvalue * 86400000
    raise ConversionError


def convert_to_seconds(string):
    if string is None or type(string) in (float, int):
        return string
    milliseconds = convert_to_milliseconds(string, default_unit='s')
    return milliseconds / 1000


def convert_to_minutes(string):
    if string is None or type(string) in (float, int):
        return string
    milliseconds = convert_to_milliseconds(string, default_unit='m')
    return milliseconds / 60000


def convert_to_hours(string):
    if string is None or type(string) in (float, int):
        return string
    milliseconds = convert_to_milliseconds(string, default_unit='h')
    return milliseconds / 3600000


def convert_to_days(string):
    if string is None or type(string) in (float, int):
        return string
    milliseconds = convert_to_milliseconds(string, default_unit='d')
    return milliseconds / 86400000


def convert_to_bytes(string, default_unit='b'):
    if string is None or type(string) in (int, ):
        return string
    string = string.lower()
    match = size_string_regex.match(string)
    if match is None:
        raise ConversionError
    svalue = float(match.group(1))
    sunit = match.group(2)
    if sunit is None:
        sunit = default_unit
    if sunit == 'b':
        return svalue
    elif sunit == 'k':
        return svalue * 1024
    elif sunit == 'm':
        return svalue * 1048576
    elif sunit == 'g':
        return svalue * 1073741824
    elif sunit == 't':
        return svalue * 1099511627776
    raise ConversionError


def convert_to_kilobytes(string):
    if string is None or type(string) in (int, ):
        return string
    bytes = convert_to_bytes(string, default_unit='k')
    return bytes / 1024


def convert_to_megabytes(string):
    if string is None or type(string) in (int, ):
        return string
    bytes = convert_to_bytes(string, default_unit='m')
    return bytes / 1048576


def convert_to_gigabytes(string):
    if string is None or type(string) in (int, ):
        return string
    bytes = convert_to_bytes(string, default_unit='g')
    return bytes / 1073741824


def convert_to_terabytes(string):
    if string is None or type(string) in (int, ):
        return string
    bytes = convert_to_bytes(string, default_unit='t')
    return bytes / 1099511627776


def convert_to_integer(string):
    if string is None or type(string) in (int, ):
        return string
    try:
        return int(string, 10)
    except ValueError:
        raise ConversionError


def convert_to_octal(string):
    if string is None or type(string) in (int, ):
        return string
    try:
        return int(string, 8)
    except ValueError:
        raise ConversionError


def convert_to_floating(string):
    if string is None or type(string) in (float, ):
        return string
    try:
        return float(string)
    except ValueError:
        raise ConversionError


def convert_to_whitespace_list(string, element_conversion=None):
    if string is None or type(string) in (list, tuple):
        return string
    result = [s.strip() for s in string.split()]
    if element_conversion is not None:
        result = [element_conversion(e) for e in result]
    return result


def convert_to_comma_list(string, element_conversion=None):
    if string is None or type(string) in (list, tuple):
        return string
    result = [s.strip() for s in string.split(',')]
    if element_conversion is not None:
        result = [element_conversion(e) for e in result]
    return result


convert_to_list = convert_to_comma_list


def convert_to_uid(string):
    if string is None or type(string) in (int, ):
        return string
    try:
        return int(string, 10)
    except ValueError:
        from pwd import getpwnam
        try:
            return getpwnam(string).pw_uid
        except KeyError:
            raise ConversionError('%s is not a valid user name' % string)


def convert_to_gid(string):
    if string is None or type(string) in (int, ):
        return string
    try:
        return int(string, 10)
    except ValueError:
        from grp import getgrname
        try:
            return getgrname(string).gr_gid
        except KeyError:
            raise ConversionError('%s is not a valid group name' % string)


##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .api import CurseAPI

##############################################################################


class Cursemeta (CurseAPI):

    CURSEMETA_BASEURL = 'https://staging_cursemeta.dries007.net/api/v3/direct/'

    def _api_baseurl(self):
        return self.CURSEMETA_BASEURL

    def get_addons(self, projectIds):
        return self._getrequest('addon', [('id', i) for i in projectIds]).json()

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from .base import BaseBuilder

##############################################################################


class CurseforgeBuilder (BaseBuilder):

    build_subloc = 'curseforge'

    default_manifest = {'minecraft': {'version': None,
                                      'modLoaders': []
                                      },
                        'manifestType': 'minecraftModpack',
                        'manifestVersion': 1,
                        'overrides': 'overrides',
                        'files': []
                        }

    ##########################################################################

    def do_build(self):
        self.log.info('Copying local files ...')
        self.copy_files(self.build_location('overrides'), self.packlock.yield_clientonly_files())

        self.log.info('Generating manifest.json ...')
        manifest = self.default_manifest
        manifest['minecraft']['version'] = self.packlock.get_metadata('minecraft_version')
        manifest['minecraft']['modLoaders'].append({'id': 'forge-{}'.format(self.packlock.get_metadata('forge_version')),
                                                    'primary': True})
        manifest['name'] = self.packlock.get_metadata('title')
        manifest['version'] = self.packlock.get_metadata('version')
        manifest['author'] = ', '.join(self.packlock.get_metadata('authors'))

        html_modlist = '\ufeff<ul>\n'

        for mod in self.packlock.yield_clientonly_mods():
            modrecord = {'projectID': mod.projectId,
                         'fileID': mod.fileId,
                         'required': True
                         }
            manifest['files'].append(modrecord)
            html_modlist += '<li><a href="{}">{} (by {})</a></li>\n'.format(mod.website,
                                                                            mod.slug,
                                                                            mod.author)
        for rep in self.packlock.get_all_resourcepacks():
            reprecord = {'projectID': rep.projectId,
                         'fileID': rep.fileId,
                         'required': True
                         }
            manifest['files'].append(reprecord)
            html_modlist += '<li><a href="{}">{} (by {})</a></li>\n'.format(rep.website,
                                                                            rep.slug,
                                                                            rep.author)
        html_modlist += '</ul>\n'

        manifestcontent = json.dumps(manifest, indent=2)
        with open('{}/manifest.json'.format(self.build_location()), 'wt') as mf:
            mf.write(manifestcontent)

        with open('{}/modlist.html'.format(self.build_location()), 'wt') as mlf:
            mlf.write(html_modlist)

        pkgfilename = '{}/{}-{}.{}'.format(self.release_location(),
                                           self.packlock.get_metadata('name'),
                                           self.packlock.get_metadata('version'),
                                           self.release_extension)
        self.log.info('Packaging the modpack: {} ...'.format(pkgfilename))
        self.release_pkg(pkgfilename)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import compress

import operator
import sys

from .baseobject import BaseObject
from .config import InvalidCommandlineArgument
from .formatter import CSVFormatter, TableFormatter, JSONFormatter, YAMLFormatter, ValueFormatter, ShellFormatter

##############################################################################


class BaseOutputter (BaseObject):

    formatters = {}
    formatter_default = None

    def add_argument_group(self, parser):
        formatter_group = parser.add_argument_group(title='output formatter',
                                                    description='output formatter options')
        self._formatter_group = formatter_group
        fchoices = sorted(self.formatters.keys())
        fdefault = self.formatter_default
        if fdefault not in fchoices:
            fdefault = fchoices[0]

        formatter_group.add_argument('-f', '--format',
                                     dest='formatter',
                                     action='store',
                                     choices=fchoices,
                                     default=fdefault,
                                     help='the output format, defaults to %s' % fdefault)
        formatter_group.add_argument('-c', '--column',
                                     action='append',
                                     default=[],
                                     dest='columns',
                                     metavar='COLUMNS',
                                     help='specify the column(s) to include, can be repeated')

        for formatter in self.formatters.values():
            f = formatter()
            f.add_argument_group(parser)
        return parser

    def output(self, arguments, column_names, data):
        self.formatter = self.formatters[arguments.formatter]()
        self.produce_output(arguments, column_names, data)

    def produce_output(self, arguments, column_names, data):
        pass

    def _generate_columns_and_selector(self, arguments, column_names):
        if not arguments.columns:
            columns_to_include = column_names
            selector = None
        else:
            columns_to_include = [c for c in column_names if c in arguments.columns]
            if not columns_to_include:
                raise InvalidCommandlineArgument('No recognized column names in %s. Recognized columns are %s.'
                                                 % (str(arguments.columns), str(column_names)))

            selector = [(c in columns_to_include) for c in column_names]
        return columns_to_include, selector

    @staticmethod
    def _compress_iterable(iterable, selectors):
        return compress(iterable, selectors)

##############################################################################


class ListOutputter (BaseOutputter):

    formatters = {'table': TableFormatter,
                  'csv': CSVFormatter,
                  'json': JSONFormatter,
                  'yaml': YAMLFormatter,
                  'value': ValueFormatter,
                  }

    formatter_default = 'table'

    def add_argument_group(self, parser):
        parser = super(ListOutputter, self).add_argument_group(parser)
        group = self._formatter_group
        group.add_argument('--sort-column',
                           action='append',
                           default=[],
                           dest='sort_columns',
                           metavar='SORT_COLUMN',
                           help=("specify the column(s) to sort the data (columns specified "
                                 "first have a priority, non-existing columns are ignored), "
                                 "can be repeated"))
        return parser

    def produce_output(self, arguments, column_names, data):
        if arguments.sort_columns:
            indexes = [column_names.index(c) for c in arguments.sort_columns if c in column_names]
            if indexes:
                data = sorted(data, key=operator.itemgetter(*indexes))
        (columns_to_include, selector) = self._generate_columns_and_selector(arguments, column_names)
        if selector:
            data = (list(self._compress_iterable(row, selector)) for row in data)

        formatter = self.formatters[arguments.formatter]()
        formatter.emit_list(columns_to_include, data, sys.stdout, arguments)

##############################################################################


class SingleOutputter (BaseOutputter):

    formatters = {'table': TableFormatter,
                  'json': JSONFormatter,
                  'shell': ShellFormatter,
                  'yaml': YAMLFormatter,
                  'value': ValueFormatter,
                  }

    formatter_default = 'table'

    def produce_output(self, arguments, column_names, data):
        (columns_to_include, selector) = self._generate_columns_and_selector(arguments, column_names)
        if selector:
            data = list(self._compress_iterable(data, selector))

        formatter = self.formatters[arguments.formatter]()
        formatter.emit_one(columns_to_include, data, sys.stdout, arguments)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import yaml
try:
    yamlload = yaml.full_load
except AttributeError:
    yamlload = yaml.load

from .framew.application import OperationError
from .packlock import PackLock

##############################################################################


class PackDefinition (object):

    def __init__(self, filenames):
        super(PackDefinition, self).__init__()
        self.definition_filenames = filenames
        self.name = None
        self.title = None
        self.version = None
        self.authors = []
        self.icon = None
        self.news = None
        self.minecraft_version = None
        self.forge_version = None
        self.routhio = None
        self.mods = {}
        self.resourcepacks = {}
        self.files = []

    def load(self):
        for filename in self.definition_filenames:
            self.load_one(filename)

    def load_one(self, filename):
        try:
            with open(filename, 'r') as pf:
                packdict = yamlload(pf)
        except FileNotFoundError:
            raise OperationError('Cannot find pack definition file: {}'.format(filename))
        except yaml.scanner.ScannerError as err:
            raise OperationError('Invalid pack definition file: {}'.format(err))
        self.populate(packdict)

    def populate(self, packdict):
        if 'name' in packdict:
            self.name = packdict['name']
        if 'title' in packdict:
            self.title = packdict['title']
        if 'version' in packdict:
            self.version = packdict['version']
        if 'authors' in packdict:
            self.authors.extend(packdict['authors'])
        if 'icon' in packdict:
            self.icon = packdict['icon']
        if 'news' in packdict:
            self.news = packdict['news']
        if 'minecraft' in packdict:
            self.minecraft_version = packdict['minecraft']
        if 'forge' in packdict:
            self.forge_version = packdict['forge']
        if 'routhio' in packdict:
            self.routhio = packdict['routhio']

        if 'mods' in packdict:
            for mod in packdict['mods']:
                if type(mod) is str:
                    slug, moddef = mod, None
                else:
                    slug, moddef = mod.popitem()
                self.mods[slug] = ModDefinition(slug, moddef)

        if 'resourcepacks' in packdict:
            for resp in packdict['resourcepacks']:
                if type(resp) is str:
                    slug, respdef = resp, None
                else:
                    slug, respdef = resp.popitem()
                self.resourcepacks[slug] = ResourcepackDefinition(slug, respdef)

        if 'files' in packdict:
            self.files.extend(packdict['files'])

    def get_mod(self, slug):
        return self.mods[slug]

    def get_all_mods(self):
        return self.mods.values()

    def get_resourcepack(self, slug):
        return self.resourcepacks[slug]

    def get_all_resourcepacks(self):
        return self.resourcepacks.values()

    def get_packlock(self):
        lockfile = os.path.splitext(self.definition_filenames[0])[0] + '.lock'
        return PackLock(lockfile)

##############################################################################


class PackElementDefinition (object):

    default_version = 'latest'
    default_optional = False
    default_recommendation = 'starred'
    default_selected = False

    def __init__(self, slug, elem):
        super(PackElementDefinition, self).__init__()
        self.slug = slug
        self.populate(elem)

    def populate(self, elem):
        if elem is not None:
            self.version = elem.get('version', self.default_version)
            self.optional = elem.get('optional', self.default_optional)
            self.recommendation = elem.get('recommendation', self.default_recommendation)
            self.selected = elem.get('selected', self.default_selected)
        else:
            self.version = self.default_version
            self.optional = self.default_optional
            self.recommendation = self.default_recommendation
            self.selected = self.default_selected

        if self.recommendation not in ('starred', 'avoid'):
            raise OperationError('addon recommendation for "{}" must be either "starred" or "avoid"'.format(self.slug))


class ModDefinition (PackElementDefinition):

    default_clientonly = False
    default_serveronly = False

    def __init__(self, slug, moddef):
        super(ModDefinition, self).__init__(slug, moddef)

    def populate(self, moddef):
        super(ModDefinition, self).populate(moddef)
        if moddef is not None:
            self.clientonly = moddef.get('clientonly', ModDefinition.default_clientonly)
            self.serveronly = moddef.get('serveronly', ModDefinition.default_serveronly)
        else:
            self.clientonly = ModDefinition.default_clientonly
            self.serveronly = ModDefinition.default_serveronly


class ResourcepackDefinition (PackElementDefinition):

    def __init__(self, slug, respdef):
        super(ResourcepackDefinition, self).__init__(slug, respdef)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['Application', 'get_app']

import argparse
import sys
import traceback

from .baseobject import BaseObject, NonStdlibError
from .config import Config, ConfigError
from .log import configure as configurelog, getlog

#############################################################################


class ApplicationError (NonStdlibError):
    """Base exception for application errors"""
    pass


class OperationError (ApplicationError):
    """Raised when a runtime application operation fails"""
    pass


#############################################################################

_app_inst = None
_app_name = "UnamedApp"
_app_version = "0"


def get_app():
    global _app_inst
    return _app_inst


def _except_hook(exc_type, value, tb):
    global _app_name, _app_version

    try:
        log = getlog()
        log.critical('\n\nException: please include the following information '
                     'in any bug report:')
        log.critical('  %s version %s' % (_app_name, _app_version))
        log.critical('  Python version %s' % sys.version.split('\n', 1)[0])
        log.critical('')
        log.critical('Unhandled exception follows:')
        tblist = (traceback.format_tb(tb, None) +
                  traceback.format_exception_only(exc_type, value))
        if type(tblist) != list:
            tblist = [tblist, ]
        for lines in tblist:
            for line in lines.split('\n'):
                if not line:
                    continue
                log.critical('%s' % line.rstrip())
        log.critical('')
        log.critical('Please also include configuration information from '
                     'running %s' % _app_name)
        log.critical('with your normal options plus \'--dump\'')
    except: # noqa
        traceback.print_exception(exc_type, value, tb)


class Application (BaseObject):

    app_name = "UnnamedApp"
    app_version = "0"
    app_description = ""

    options = {'exception_hook': _except_hook,
               'argparse_kwargs': None,
               }

    def __init__(self, **kw):
        """constructor"""
        super(Application, self).__init__()
        self._parse_options(Application.options, kw)

        if hasattr(self, 'config_files'):
            self.app_config = self.config_files
        elif hasattr(self, 'config_file'):
            self.app_config = str(self.config_file)
        else:
            self.app_config = "%s.conf" % self.app_name

        global _app_inst, _app_name, _app_version
        _app_inst = self
        _app_name = self.app_name
        _app_version = self.app_version

        if self.exception_hook is not None:
            sys.excepthook = self.exception_hook

        self.cmdline_parser = self.build_cmdline_parser()
        self.additional_arguments = None

    def build_cmdline_parser(self):
        argparse_kwargs = self.argparse_kwargs or {}
        parser = argparse.ArgumentParser(description=self.app_description, **argparse_kwargs)
        parser.add_argument('--version',
                            action='version',
                            version='{0} {1}'.format(self.app_name, self.app_version))
        verbose_group = parser.add_mutually_exclusive_group()
        verbose_group.add_argument('-v', '--verbose',
                                   action='count',
                                   dest='verbose_level',
                                   default=1,
                                   help='Increase verbosity of output. Can be repeated.')
        verbose_group.add_argument('-q', '--quiet',
                                   action='store_const',
                                   dest='verbose_level',
                                   const=0,
                                   help='Supress output except warnings and errors.')
        parser.add_argument('--config',
                            action='store',
                            dest='app_config',
                            help='Specify an alternative config file to read.')
        parser.add_argument('--logfile',
                            action='store',
                            default=None,
                            help='Specify a file to log output. Disabled by default.')
        parser.add_argument('--debug',
                            default=False,
                            action='store_true',
                            help='Show tracebacks on errors.')
        parser.add_argument('--dump',
                            action='store_true',
                            help='Output configuration and other debugging information')
        return parser

    def start(self):
        try:
            self.cmdline_arguments, self.additional_arguments = self.cmdline_parser.parse_known_args(sys.argv[1:])
            if self.cmdline_arguments.app_config is not None:
                self.app_config = self.cmdline_arguments.app_config
            self.config = Config(self.app_config)
            if self.cmdline_arguments.dump:
                self.dump()

            configurelog(self.config, self.cmdline_arguments)

            self.setup(self.additional_arguments)
            self.run()
            self.cleanup()
            return 0

        except KeyboardInterrupt:
            self.abort('Aborted by user (keyboard interrupt)', errorcode=0)

        except ConfigError as o:
            self.abort('Configuration error: %s' % (o), errorcode=2)

        except OperationError as o:
            self.abort('Error: %s' % (o), errorcode=3)

        # otherwise, let the exception_hook handle the error (if defined)

    def setup(self, unknown_arguments):
        pass

    def cleanup(self):
        pass

    def run(self):
        raise NotImplementedError("Override this function to do something")

    def dump(self):
        sys.stdout.write('#########################################################\n')
        sys.stdout.write('# PACKMAKER DEBUG INFO DUMP:\n')
        sys.stdout.write('#    cmdline_arguments = {}\n'.format(self.cmdline_arguments))
        sys.stdout.write('#    additional_arguments = {}\n'.format(self.additional_arguments))
        sys.stdout.write('#    config filenames = {}\n'.format(self.config.filenames))
        sys.stdout.write('#    config contents:\n')
        self.config.dump_config(sys.stdout)
        sys.stdout.write('#########################################################\n')

    def abort(self, message, errorcode=1):
        sys.stderr.write('%s\n' % message)
        sys.stderr.flush()
        try:
            self.cleanup()
        except Exception as why:
            sys.stderr.write('UNEXPECTED ERROR during cleanup: %s\n' & str(why))
            sys.stderr.flush()
        sys.exit(errorcode)

#############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import dateutil.parser
import json

from ..curse.curseforge import Curseforge
from ..framew.application import OperationError
from ..framew.cmdapplication import Subcommand
from ..framew.config import ConfigError
from ..framew.log import getlog
from ..packdef import PackDefinition

##############################################################################


class LockCommand (Subcommand):
    """
    Lock the modpack. Find mod download urls, generate a packmaker.lock file.
    """

    name = 'lock'

    default_packmaker_yml = 'packmaker.yml'

    def setup(self):
        super(LockCommand, self).setup()
        self.log = getlog()
        self.setup_api()
        self.setup_db()

    def setup_api(self):
        authn_token = self.config.get('curseforge::authentication_token')
        if authn_token is None:
            raise ConfigError('No curseforge authentication token')
        self.api = Curseforge(authn_token)

    def setup_db(self):
        db_filename = self.config.get('curseforge::moddb_filename')
        if db_filename is None:
            raise ConfigError('No moddb_filename parameter in configuration')

        self.log.info('Loading curseforge database ...')
        with open(db_filename, 'r') as cf:
            self.db = json.load(cf)

    def get_cmdline_parser(self):
        parser = super(LockCommand, self).get_cmdline_parser()
        parser.add_argument('packdef',
                            nargs='*',
                            default=[LockCommand.default_packmaker_yml],
                            help='modpack definition file')
        return parser

    def run_command(self, parsed_args):

        self.log.info('Reading pack definition ...')
        pack = PackDefinition(parsed_args.packdef)
        pack.load()

        packlock = pack.get_packlock()
        packlock.set_all_metadata(pack)

        self.resolve_mods(pack, packlock)
        self.resolve_resourcepacks(pack, packlock)

        self.log.info('Adding files...')
        for filesdef in pack.files:
            packlock.add_files(filesdef)

        self.log.info('Adding extra options (if any)...')
        if pack.routhio is not None:
            packlock.add_extraopt('routhio', pack.routhio)

        self.log.info('Writing pack lock...')
        packlock.save()

        self.log.info('Done.')

    def resolve_mods(self, pack, packlock):
        self.log.info('Resolving mods...')
        for moddef in pack.get_all_mods():
            if moddef.slug in self.db and self.db[moddef.slug]['category'] == 'mod':
                modid = self.db[moddef.slug]['id']
                modname = self.db[moddef.slug]['name']
                modauthor = self.db[moddef.slug]['authors'][0]['name']
                modwebsite = self.db[moddef.slug]['websiteUrl']
            else:
                modname, modid, modauthor, modwebsite = self.manual_addon_search(moddef, 6, pack.minecraft_version)

            modfile_found = self.resolve_addon(modid, moddef, pack.minecraft_version)
            if modfile_found is None:
                raise OperationError('Cannot find a mod file for {}'.format(moddef.slug))

            packlock.add_resolved_mod(moddef, {'projectId': modid,
                                               'name': modname,
                                               'author': modauthor,
                                               'website': modwebsite,
                                               'fileId': modfile_found['id'],
                                               'fileName': modfile_found['fileName'],
                                               'downloadUrl': modfile_found['downloadUrl']
                                               })

    def resolve_resourcepacks(self, pack, packlock):
        self.log.info('Resolving resourcepacks...')
        for repdef in pack.get_all_resourcepacks():
            if repdef.slug in self.db and self.db[repdef.slug]['category'] == 'resourcepack':
                repid = self.db[repdef.slug]['id']
                repname = self.db[repdef.slug]['name']
                repauthor = self.db[repdef.slug]['authors'][0]['name']
                repwebsite = self.db[repdef.slug]['websiteUrl']
            else:
                repname, repid, repauthor, repwebsite = self.manual_addon_search(repdef, 12, pack.minecraft_version)

            repfile_found = self.resolve_addon(repid, repdef, pack.minecraft_version)
            if repfile_found is None:
                raise OperationError('Cannot find a resourcepack file for {}'.format(repdef.slug))

            packlock.add_resolved_resourcepack(repdef, {'projectId': repid,
                                                        'name': repname,
                                                        'author': repauthor,
                                                        'website': repwebsite,
                                                        'fileId': repfile_found['id'],
                                                        'fileName': repfile_found['fileName'],
                                                        'downloadUrl': repfile_found['downloadUrl']
                                                        })

    def manual_addon_search(self, addondef, addontype, gameversion):
        self.log.info('  Cannot find addon in local db: {}\n  Manually searching...'.format(addondef.slug))

        searchresults = list(({'name': addon['name'], 'id': addon['id'], 'slug': addon['slug'],
                               'authors': addon['authors'], 'websiteUrl': addon['websiteUrl']}
                              for addon in self.api.yield_addons_by_criteria(gameId=432, sectionId=addontype,
                                                                             gameVersions=gameversion,
                                                                             searchFilter=addondef.slug)))

        if len(searchresults) < 1:
            raise OperationError('Cannot find an addon named \'{}\''.format(addondef.slug))
        elif len(searchresults) > 1:
            self.log.info('    Multiple search results found ({}).  Looking for an exact match in results...'
                          .format(len(searchresults)))
            for sresult in searchresults:
                if sresult['slug'] == addondef.slug:
                    searchresult = sresult
                    self.log.info('    Found it! ... {} (id = {})'.format(searchresult['slug'], searchresult['id']))
                    break
            else:
                searchresult = searchresults[0]
                self.log.info('    No exact match found! Using the first one (this could be wildly wrong) ... {} (id = {})'
                              .format(searchresult['slug'], searchresult['id']))
        else:
            searchresult = searchresults[0]
            self.log.info('    Found it! ... {} (id = {})'.format(searchresult['slug'], searchresult['id']))

        return (searchresult['name'], searchresult['id'], searchresult['authors'][0]['name'], searchresult['websiteUrl'])

    def resolve_addon(self, id, addondef, gameversion):
        latestTimestamp = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
        addon_found = None
        for addonfile in self.api.get_addon_files(id):
            if gameversion not in addonfile['gameVersion']:
                continue
            if addondef.version == 'latest':
                timestamp = dateutil.parser.parse(addonfile['fileDate'])
                if timestamp > latestTimestamp:
                    addon_found = addonfile
                    latestTimestamp = timestamp

            elif addonfile['fileName'] == addondef.version:
                addon_found = addonfile

        return addon_found

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ..builder.curseforge import CurseforgeBuilder
from ..builder.local import LocalBuilder
from ..builder.server import ServerBuilder
from ..builder.routhio import RouthioBuilder
from ..framew.cmdapplication import Subcommand

##############################################################################


class BuildBase (Subcommand):
    builderclass = None

    ##########################################################################

    def setup(self):
        super(BuildBase, self).setup()
        self.builder = self.builderclass(self.config)

    ##########################################################################

    def get_cmdline_parser(self):
        parser = super(BuildBase, self).get_cmdline_parser()
        self.builder.add_cmdline_args(parser)
        return parser

    ##########################################################################

    def run_command(self, parsed_args):
        self.builder.setup_build(parsed_args)
        self.builder.do_build()

##############################################################################


class BuildCurseforge (BuildBase):
    """
    Build a curseforge compatible modpack.
    """
    name = 'build-curseforge'
    builderclass = CurseforgeBuilder

##############################################################################


class BuildServer (BuildBase):
    """
    Build a server modpack.
    """
    name = 'build-server'
    builderclass = ServerBuilder

##############################################################################


class BuildRouthio (BuildBase):
    """
    Build a Routh.io compatible modpack.
    """
    name = 'build-routhio'
    builderclass = RouthioBuilder

##############################################################################


class BuildLocal (BuildBase):
    """
    Build a local installation
    """
    name = 'build-local'
    builderclass = LocalBuilder

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import ArgumentParser, Action

import inspect
import sys

from .application import Application, OperationError
from .baseobject import BaseObject

##############################################################################


class CommandApplication (Application):

    subcommands = []

    def __init__(self, **kw):
        kw['argparse_kwargs'] = {'add_help': False}
        super(CommandApplication, self).__init__(**kw)
        self.command = None

    def build_cmdline_parser(self):
        parser = super(CommandApplication, self).build_cmdline_parser()
        parser.add_argument('-h', '--help',
                            dest='deferred_help',
                            action='store_true',
                            help='Show help message and exit.')
        return parser

    def setup(self, subcommand_arguments):
        self.setup_help(subcommand_arguments)
        self.setup_subcommand(subcommand_arguments)

    def setup_help(self, arguments):
        if self.cmdline_arguments.deferred_help and arguments:
            self.cmdline_arguments.deferred_help = False
            arguments.insert(0, "help")

        self.subcommands.append(HelpCommand)

        if self.cmdline_arguments.deferred_help:
            action = HelpAction(None, None, default=self)
            action(self.cmdline_parser, self.cmdline_arguments, None, None)

    def setup_subcommand(self, arguments):
        if not arguments:
            action = HelpAction(None, None, default=self)
            action(self.cmdline_parser, self.cmdline_arguments, None, None)

        command = arguments[0]
        command_args = arguments[1:]

        self.command = self._find_subcommand(command)

        cmd_parser = self.command.get_cmdline_parser()
        self.arguments = cmd_parser.parse_args(command_args)

    def _find_subcommand(self, command):
        for cmdclass in self.subcommands:
            if cmdclass.name is None:
                continue
            if command == cmdclass.name:
                return cmdclass(self)
        else:
            raise OperationError('unknown command: %s' % command)

    def run(self):
        self.command.run(self.arguments)

##############################################################################


class HelpAction (Action):

    def __call__(self, parser, namespace, values, option_string=None):
        app = self.default
        parser.print_help(sys.stdout)
        sys.stdout.write('\nCommands:\n')

        for cmdclass in app.subcommands:
            if cmdclass.name is None:
                continue

            cmd = cmdclass(app)
            cmdname = cmd.name
            cmddesc = cmd.get_description().split('\n')[0]

            sys.stdout.write('  %-19s  %s\n' % (cmdname, cmddesc))

        sys.exit(0)

##############################################################################


class Subcommand (BaseObject):

    name = None
    description = None

    def __init__(self, app, **kw):
        super(Subcommand, self).__init__(**kw)
        self.app = app
        self.setup()

    def setup(self):
        self.config = self.app.config

    def get_description(self):
        if self.description is None:
            return inspect.getdoc(self.__class__) or ''
        return self.description

    def get_cmdline_parser(self):
        parser = ArgumentParser(description=self.get_description(),
                                prog='%s %s' % (self.app.app_name, self.name))
        return parser

    def run(self, arguments):
        self.run_command(arguments)

    def run_command(self, arguments):
        pass

##############################################################################


class HelpCommand (Subcommand):
    """
    Print detailed help for another command
    """

    name = 'help'

    def get_cmdline_parser(self):
        parser = super(HelpCommand, self).get_cmdline_parser()
        parser.add_argument('cmd',
                            help='name of the command')
        return parser

    def run_command(self, arguments):
        if arguments.cmd:
            thecmd = self.app._find_subcommand(arguments.cmd)
            parser = thecmd.get_cmdline_parser()
            parser.print_help(sys.stdout)
        else:
            action = HelpAction(None, None, default=self.app)
            action(self.cmdline_parser, self.cmdline_arguments, None, None)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A configuration file parser.
"""

import os
import os.path
import re

from .baseobject import BaseObject, NonStdlibError
from .string_conversion import convert_to_bool, convert_to_list

##############################################################################


class ConfigError (NonStdlibError):
    """Base exception class for Config exceptions"""
    pass


class InvalidCommandlineArgument (ConfigError):
    pass


class InvalidConfigParameter (ConfigError):
    pass

##############################################################################


class Config (BaseObject):

    DEFAULT_CONFIG_PATHS = ['.', '${HOME}', '${HOME}/.config', '/usr/local/etc', '/etc']

    SECTION_REX = re.compile(r'^\[(.*?)\]')
    KEYVALUE_REX = re.compile(r'^([^=]+)=(.*)$')
    SUBST_REX = re.compile(r'\${([\w:]+)}')
    VARSEC_REX = re.compile('^([a-z]+)::(.*)')

    options = {'base_config_dict':              None,
               'error_on_invalid_config_param': (False, convert_to_bool),
               'skipped_sections':              ([], convert_to_list),
               }

    def __init__(self, filenames=[], **kw):
        """constructor"""
        super(Config, self).__init__(**kw)
        self._parse_options(Config.options, kw)

        if type(filenames) != list:
            filenames = [filenames, ]
        self.filenames = filenames
        self.conf = dict()
        self.cmd_conf = dict()

        if self.filenames:
            self._load_config_files()
        else:
            if self.base_config_dict is None:
                self.conf = dict(DEFAULT=dict())
            else:
                self.conf = dict(DEFAULT=dict(self.base_config_dict))

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except LookupError:
            return default

    def get_sections(self):
        sections = list(self.conf)
        sections.sort()
        return sections

    def section(self, section):
        return self.conf[section]

    def dump_config(self, file):
        sections = self.conf.keys()
        for sec in sections:
            if type(self.conf[sec]) != dict:
                continue
            file.write('[%s]\n' % sec)
            params = self.conf[sec].keys()
            for param in params:
                file.write('%s = %s\n' % (param, self.conf[sec][param]))
            file.write('\n')

    def read_nonconfig_section(self, section_name, concat_all_config_files=False):
        rev_filenames = self.filenames
        rev_filenames.reverse()
        found_section = False
        section_lines = []
        for filename in rev_filenames:
            config_filename = self._find_config_file(filename)
            if config_filename is None:
                continue

            fp = open(config_filename, 'r')
            lines = fp.readlines()
            fp.close()

            linecount = 0
            num_lines = len(lines)

            # find the section_name
            while linecount < num_lines:
                line = lines[linecount].strip()
                linecount += 1
                if line == '[%s]' % section_name:
                    found_section = True
                    break
            if not found_section:
                continue

            # read the lines until the next section or eof
            while linecount < num_lines:
                line = lines[linecount].strip()
                linecount += 1
                if line and line[0] == '[' and line[-1] == ']':
                    break
                section_lines.append(line)
            linecount += 1

            if found_section and concat_all_config_files is False:
                return section_lines

        return section_lines

    def _load_config_files(self):
        if self.base_config_dict is None:
            self.conf = dict(DEFAULT=dict())
        else:
            self.conf = dict(DEFAULT=dict(self.base_config_dict))
        for filename in self.filenames:
            configFilename = self._find_config_file(filename)
            if configFilename is None:
                continue

            fp = open(configFilename, 'r')
            lines = fp.readlines()
            fp.close()

            config_lines = []
            linecount = 0
            while linecount < len(lines):
                line = re.sub('#.*$', '', lines[linecount]).strip()
                if not line:
                    linecount += 1
                    continue
                while line[-1] == '\\':
                    line = line[:-1] + lines[linecount+1].strip()
                    linecount += 1
                config_lines.append(line)
                linecount += 1

            section = 'DEFAULT'
            in_skipped_section = False
            for line in config_lines:
                section_mo = Config.SECTION_REX.match(line)
                if section_mo:
                    section = section_mo.group(1)
                    if section in self.skipped_sections:
                        in_skipped_section = True
                    else:
                        in_skipped_section = False
                        self.conf[section] = dict()
                    continue

                if in_skipped_section is True:
                    continue

                if Config.KEYVALUE_REX.match(line):
                    key, value = line.split('=', 1)
                    key = key.rstrip()
                    value = value.lstrip()

                    value = os.path.expanduser(os.path.expandvars(value))

                    # Check for variable substitution
                    # This allows you to stick things like "${key}" or "${section::key}"
                    # into the config file.
                    subst = Config.SUBST_REX.search(value)
                    while subst:
                        var = subst.group(1)
                        # Work around the use of ${} for other purposes (for now)
                        varsec = Config.VARSEC_REX.match(var)
                        if varsec:
                            prefix = self.conf[varsec.group(1)][varsec.group(2)] or ""
                        else:
                            prefix = self.conf[section][var] or ""
                        # If not subst was found, check for an environment variable
                        if not prefix:
                            try:
                                prefix = os.environ[var]
                            except KeyError:
                                pass
                        value = re.sub(r'\${%s}' % var, prefix, value)
                        subst = Config.SUBST_REX.search(value)

                    self.conf[section][key] = value

                elif self.error_on_invalid_config_param is True:
                    raise InvalidConfigParameter("Invalid config parameter: %s" % line)

                # else ignore any invalid lines

    def _find_config_file(self, filename):
        if filename.startswith("/"):
            if os.path.exists(filename):
                return filename
        else:
            for path in Config.DEFAULT_CONFIG_PATHS:
                fullname = os.path.expandvars(os.path.expanduser(os.path.join(path, filename)))
                if os.path.exists(fullname):
                    return fullname
        return None

    def __getitem__(self, key):
        if key in self.cmd_conf:
            return self.cmd_conf[key]
        else:
            if key.find('::') >= 0:
                section, key = key.split('::', 1)
                if not section:
                    section = 'DEFAULT'
            else:
                section = 'DEFAULT'
            return self.conf[section][key]


##############################################################################

if __name__ == "__main__":
    pass

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess

from .build import BuildLocal
from ..forge import Forge
from ..framew.log import getlog
from ..minecraft import Minecraft

##############################################################################


class LaunchCommand (BuildLocal):
    """
    Launch a local installation
    """
    name = 'launch'

    default_java_binary = 'java'
    default_java_memory = '4096m'
    default_java_arguments = '-XX:+UseG1GC ' \
                             '-XX:+UnlockExperimentalVMOptions ' \
                             '-XX:G1NewSizePercent=20 ' \
                             '-XX:G1ReservePercent=20 ' \
                             '-XX:MaxGCPauseMillis=50 ' \
                             '-XX:G1HeapRegionSize=32M ' \
                             '-Dsun.rmi.dgc.server.gcInterval=2147483646 ' \
                             '-Dfml.readTimeout=180'

    def setup(self):
        super(LaunchCommand, self).setup()
        self.setup_java_arguments()

    def setup_java_arguments(self):
        self.java_binary = self.config.get('launch::java_binary', self.default_java_binary)
        self.java_memory = self.config.get('launch::java_memory', self.default_java_memory)
        self.java_arguments = self.config.get('launch::java_arguments', self.default_java_arguments)

    def run_command(self, parsed_args):
        super(LaunchCommand, self).run_command(parsed_args)
        self.run_launcher()

    def run_launcher(self):
        mc = Minecraft(self.builder.packlock.get_metadata('minecraft_version'))
        launch_args = mc.get_launch_arguments()
        main_class = mc.get_mainclass()

        forge = None
        if self.builder.packlock.get_metadata('forge_version') is not None:
            forge = Forge(self.builder.packlock.get_metadata('minecraft_version'),
                          self.builder.packlock.get_metadata('forge_version'),
                          self.builder.downloader.subdownloader('forge'))

            forge_launch_args = forge.get_launch_arguments()

            if forge_launch_args.startswith('--username ${auth_player_name}'):
                launch_args = forge_launch_args
            else:
                launch_args = '{} {}'.format(launch_args, forge_launch_args)

            main_class = forge.get_mainclass()

        manifest = mc.get_version_manifest()

        classpath = [os.path.join('libraries', 'com', 'mojang', 'minecraft', manifest['id'],
                                  'minecraft-{}-client.jar'.format(manifest['id']))]

        libraries = {}

        for library in mc.get_libraries():
            if library.must_extract():
                continue
            libraries[library.get_artifact()] = os.path.join('libraries', library.get_path())

        if forge is not None:
            for library in forge.get_dependent_libraries():
                libraries[library.get_artifact()] = os.path.join('libraries', library.get_path())

        classpath.extend(libraries.values())
        classpath = ':'.join(classpath)

        launch_args = launch_args.replace('${auth_player_name}', 'Player')
        launch_args = launch_args.replace('${version_name}', manifest['id'])
        launch_args = launch_args.replace('${game_directory}', '.')
        launch_args = launch_args.replace('${assets_root}', './assets')
        launch_args = launch_args.replace('${assets_index_name}', manifest['assets'])
        launch_args = launch_args.replace('${auth_uuid}', '00000000-0000-0000-0000-000000000000')
        launch_args = launch_args.replace('${auth_access_token}', '0')
        launch_args = launch_args.replace('${user_type}', 'mojang')
        launch_args = launch_args.replace('${version_type}', manifest['type'])

        cmdline = [self.java_binary, '-cp', classpath,
                   '-Xms{}'.format(self.java_memory), '-Xmx{}'.format(self.java_memory),
                   self.java_arguments,
                   '-Djava.library.path={}'.format('./natives'),
                   main_class, launch_args]

        print('DEBUG: cmd = {}'.format(' '.join(cmdline)))

        result = subprocess.run(' '.join(cmdline), cwd=self.builder.build_location(), shell=True)
        if result.returncode != 0:
            getlog().warn('Minecraft process exited with an error.')

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from ..curse.curseforge import Curseforge
from ..framew.cmdapplication import Subcommand
from ..framew.config import ConfigError
from ..framew.log import getlog
from ..framew.string_conversion import convert_to_bool

##############################################################################


class UpdateDbCommand (Subcommand):
    """
    Download and compile a new mods database from curseforge.
    """
    name = 'updatedb'

    batchsz = 1000

    addon_categories = {'mod': 6, 'resourcepack': 12}

    def get_cmdline_parser(self):
        parser = super(UpdateDbCommand, self).get_cmdline_parser()
        parser.add_argument('--ignore-mods', action='store_true',
                            help='Do not scan for mods when updating the curseforge db')
        parser.add_argument('--ignore-resourcepacks', action='store_true',
                            help='Do not scan for resourcepacks when updating the curseforge db')
        return parser

    def setup(self):
        super(UpdateDbCommand, self).setup()
        self.setup_api()
        self.setup_db()
        self.setup_ignores()
        self.log = getlog()

    def setup_api(self):
        authn_token = self.config.get('curseforge::authentication_token')
        if authn_token is None:
            raise ConfigError('No curseforge authentication token')
        self.api = Curseforge(authn_token)

    def setup_db(self):
        db_filename = self.config.get('curseforge::moddb_filename')
        if db_filename is None:
            raise ConfigError('No moddb_filename parameter in configuration')
        self.db_filename = db_filename

    def setup_ignores(self):
        self.ignore_mods = convert_to_bool(self.config.get('curseforge::ignore_mods', 'false'))
        self.ignore_resourcepacks = convert_to_bool(self.config.get('curseforge::ignore_resourcepacks', 'false'))

    def run_command(self, parsed_args):
        addons = {}

        dbdir = os.path.dirname(self.db_filename)
        if dbdir not in ['', '.'] and not os.path.exists(dbdir):
            self.log.info('Creating directory: {}'.format(dbdir))
            os.makedirs(dbdir)

        if parsed_args.ignore_mods or self.ignore_mods:
            self.log.info('Not scanning for mod addons.  Configured to ignore mods.')
        else:
            self.log.info('Scanning for list of possible mod addons...')
            addons = self.get_addons_from_api('mod')

        if parsed_args.ignore_resourcepacks or self.ignore_resourcepacks:
            self.log.info('Not scanning for resourcepack addons.  Configure to ignore resourcepacks.')
        else:
            self.log.info('Scanning for list of possible resourcepack addons...')
            addons.update(self.get_addons_from_api('resourcepack'))

        self.log.info('Writing addons database...')
        db = json.dumps(addons)
        with open(self.db_filename, 'w') as dbfile:
            dbfile.write(db)

        self.log.info('Done.')

    def get_addons_from_api(self, category):
        addon_dict = {}
        processedIDs = []
        for addon in self.api.yield_addons_by_criteria(gameId=432,
                                                       sectionId=self.addon_categories[category],
                                                       sort='LastUpdated',
                                                       pageSize=UpdateDbCommand.batchsz):
            processedIDs.append(addon['id'])

        self.log.info('  Reading individual addon information...')
        chunks = [processedIDs[i:i + UpdateDbCommand.batchsz] for i in range(0, len(processedIDs), UpdateDbCommand.batchsz)]
        for chunkIDs in chunks:
            addons = self.api.get_addons(chunkIDs)
            for addon in addons:
                if not addon:
                    continue
                addon_dict[addon['slug']] = {'id': addon['id'],
                                             'name': addon['name'],
                                             'slug': addon['slug'],
                                             'category': category,
                                             'authors': addon['authors'],
                                             'websiteUrl': addon['websiteUrl']}
        return addon_dict

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['CSVFormatter', 'TableFormatter', ]

import argparse
import csv
import json
import os
import prettytable
import yaml

from .baseobject import BaseObject

##############################################################################


class BaseFormatter (BaseObject):
    def add_argument_group(self, parser):
        pass


class ListFormatter (BaseFormatter):
    def emit_list(self, column_names, data, stdout, arguments):
        pass


class SingleFormatter (BaseFormatter):
    def emit_one(self, column_names, data, stdout, arguments):
        pass

##############################################################################


class CSVFormatter (ListFormatter):

    QUOTE_MODES = {'all': csv.QUOTE_ALL,
                   'minimal': csv.QUOTE_MINIMAL,
                   'nonnumeric': csv.QUOTE_NONNUMERIC,
                   'none': csv.QUOTE_NONE
                   }

    def add_argument_group(self, parser):
        group = parser.add_argument_group('CSV formatter')
        group.add_argument('--quote',
                           choices=sorted(self.QUOTE_MODES.keys()),
                           dest='quote_mode',
                           default='nonnumeric',
                           help='when to include quotes, default to nonnumeric')

    def emit_list(self, column_names, data, stdout, arguments):
        writer_kwargs = dict(quoting=self.QUOTE_MODES[arguments.quote_mode],
                             lineterminator=os.linesep,
                             escapechar='\\')

        writer = csv.writer(stdout, **writer_kwargs)
        writer.writerow(column_names)
        for row in data:
            writer.writerow([c for c in row])

##############################################################################


class TableFormatter (ListFormatter, SingleFormatter):

    ALIGNMENTS = {int: 'r',
                  str: 'l',
                  float: 'r'
                  }
    try:
        ALIGNMENTS[unicode] = 'l'
    except NameError:
        pass

    def add_argument_group(self, parser):
        group = parser.add_argument_group('Table formatter')
        group.add_argument('--max-width',
                           metavar='<integer>',
                           default=int(os.environ.get('MAX_TERM_WIDTH', 0)),
                           type=int,
                           help=('maximum display width, <1 to disable. You cal also '
                                 'use the MAX_DISPLAY_WIDTH environment variable, '
                                 'but the parameter takes precedence.'))
        group.add_argument('--fit-width',
                           action='store_true',
                           default=bool(int(os.environ.get('FIT_WIDTH', 0))),
                           help=('Fit the table to the display width. '
                                 'Implied if --max-width greater than 0. '
                                 'Set the environment variable FIT_WIDTH=1 '
                                 'to always enable'))
        group.add_argument('--print-empty',
                           action='store_true',
                           help='Print empty table iof there is no data to show.')

    def emit_list(self, column_names, data, stdout, arguments):
        x = prettytable.PrettyTable(column_names, print_empty=arguments.print_empty)
        x.padding_width = 1

        if data:
            self.add_rows(x, column_names, data)

        min_width = 0
        self._assign_max_widths(stdout, x, int(arguments.max_width), min_width, arguments.fit_width)

        formatted = x.get_string()
        stdout.write(formatted)
        stdout.write('\n')

    def emit_one(self, column_names, data, stdout, arguments):
        x = prettytable.PrettyTable(field_names=('Field', 'Value'),
                                    print_empty=False)
        x.padding_width = 1
        x.align['Field'] = 'l'
        x.align['Value'] = 'l'
        for name, value in zip(column_names, data):
            x.add_row(self._format_row((name, value)))

        min_width = 16
        self._assign_max_widths(stdout, x, int(arguments.max_width), min_width, arguments.fit_width)

        formatted = x.get_string()
        stdout.write(formatted)
        stdout.write('\n')

    def add_rows(self, table, column_names, data):
        data_iter = iter(data)
        try:
            first_row = next(data_iter)
        except StopIteration:
            pass
        else:
            for value, name in zip(first_row, column_names):
                alignment = self.ALIGNMENTS.get(type(value), 'l')
                table.align[name] = alignment
            table.add_row(self._format_row(first_row))
            for row in data_iter:
                table.add_row(self._format_row(row))

    @staticmethod
    def _assign_max_widths(stdout, x, max_width, min_width=0, fit_width=False):
        pass

    @staticmethod
    def _format_row(row):
        new_row = []
        for r in row:
            new_row.append(r)
        return new_row

##############################################################################


class JSONFormatter (ListFormatter, SingleFormatter):

    def add_argument_group(self, parser):
        group = parser.add_argument_group(title='json formatter')
        group.add_argument('--noindent',
                           action='store_true',
                           dest='noindent',
                           help='whether to disable indenting the JSON')

    def emit_list(self, column_names, data, stdout, arguments):
        items = []
        for item in data:
            items.append({n: (i) for n, i in zip(column_names, item)})
        indent = None if arguments.noindent else 2
        json.dump(items, stdout, indent=indent)
        stdout.write('\n')

    def emit_one(self, column_names, data, stdout, arguments):
        one = {n: (i) for n, i in zip(column_names, data)}
        indent = None if arguments.noindent else 2
        json.dump(one, stdout, indent=indent)
        stdout.write('\n')

##############################################################################


class ShellFormatter (SingleFormatter):

    def add_argument_group(self, parser):
        group = parser.add_argument_group(title='shell formatter',
                                          description='a format a UNIX shell can parse (variable="value")')
        group.add_argument('--variable',
                           action='append',
                           default=[],
                           dest='variables',
                           metavar='VARIABLE',
                           help=argparse.SUPPRESS)
        group.add_argument('--prefix',
                           action='store',
                           default='',
                           dest='prefix',
                           help='add a prefix to all variable names')

    def emit_one(self, column_names, data, stdout, arguments):
        variable_names = [c.lower().replace(' ', '_') for c in column_names]
        desired_columns = arguments.variables
        for name, value in zip(variable_names, data):
            if name in desired_columns or not desired_columns:
                value = str(value)
                value = value.replace('"', '\\"')
                name = name.replace(':', '_')
                name = name.replace('-', '_')

                stdout.write('%s%s="%s"\n' % (arguments.prefix, name, value))

##############################################################################


class ValueFormatter (ListFormatter, SingleFormatter):

    def emit_list(self, column_names, data, stdout, arguments):
        for row in data:
            stdout.write(' '.join(str(c) for c in row) + u'\n')

    def emit_one(self, column_names, data, stdout, arguments):
        for value in data:
            stdout.write('%s\n' % value)

##############################################################################


class YAMLFormatter (ListFormatter, SingleFormatter):

    def emit_list(self, column_names, data, stdout, arguments):
        items = []
        for item in data:
            items.append({n: (i) for n, i in zip(column_names, item)})
        yaml.safe_dump(items, stream=stdout, default_flow_style=False)

    def emit_one(self, column_names, data, stdout, arguments):
        for key, value in zip(column_names, data):
            dict_data = {key: (value)}
            yaml.safe_dump(dict_data, stream=stdout, default_flow_style=False)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import zipfile

from .base import BaseBuilder
from ..download import HttpDownloader
from ..forge import Forge
from ..minecraft import Minecraft

##############################################################################


class LocalBuilder (BaseBuilder):

    build_subloc = 'local'

    ##########################################################################

    def setup_build(self, parsed_args):
        super(LocalBuilder, self).setup_build(parsed_args)

        self.mods_location = self.build_location('mods')
        self.resourcepacks_location = self.build_location('resourcepacks')
        self.libs_location = self.build_location('libraries')
        self.natives_location = self.build_location('natives')
        self.assets_location = self.build_location('assets')

    ##########################################################################

    def do_build(self):
        self.downloader = HttpDownloader(self.cache_location())

        self.log.info('Copying local files ...')
        self.copy_files(files_iterator=self.packlock.yield_clientonly_files())

        self.log.info('Copying mod files ...')
        self.copy_mods(self.mods_location, self.packlock.yield_clientonly_mods())

        self.log.info('Copying resourcepack files ...')
        self.copy_resourcepacks(self.resourcepacks_location)

        mc = Minecraft(self.packlock.get_metadata('minecraft_version'))
        self.copy_minecraft_client(mc)

        if self.packlock.get_metadata('forge_version') is not None:
            self.copy_forge(mc)

    ##########################################################################

    def copy_minecraft_client(self, mc):
        mc.download_client(self.downloader.subdownloader('minecraft'))

        manifest = mc.get_version_manifest()

        # Copy the minecraft client library into place.

        cached_lib = os.path.join(self.cache_location('minecraft'), 'versions', manifest['id'], 'client.jar')
        local_lib = os.path.join(self.libs_location, 'com', 'mojang', 'minecraft', manifest['id'],
                                 'minecraft-{}-client.jar'.format(manifest['id']))

        if not os.path.exists(os.path.dirname(local_lib)):
            os.makedirs(os.path.dirname(local_lib))

        shutil.copy2(cached_lib, local_lib)

        # Copy the minecraft libraries into place.

        for lib in mc.get_libraries():
            cached_lib = os.path.join(self.cache_location('minecraft'), 'libraries', lib.get_path())

            if lib.must_extract():
                with zipfile.ZipFile(cached_lib, 'r') as extractedlib:
                    for name in extractedlib.namelist():
                        excluded = False
                        for exclude in lib.extract_excludes():
                            if name.startswith(exclude):
                                excluded = True
                                break
                        if excluded:
                            continue
                        extractedlib.extract(name, path=self.natives_location)

            else:
                localdir = os.path.join(self.libs_location, os.path.dirname(lib.get_path()))
                if not os.path.exists(localdir):
                    os.makedirs(localdir)

                local_lib = os.path.join(self.libs_location, lib.get_path())
                shutil.copy2(cached_lib, local_lib)

        # Copy the minecraft assets into place.

        assetIndexDir = os.path.join(self.assets_location, 'indexes')
        if not os.path.exists(assetIndexDir):
            os.makedirs(assetIndexDir)

        assetIndexFile = '{}.json'.format(manifest['assets'])

        cached_assetIndexFile = os.path.join(self.cache_location('minecraft'), 'assets', 'indexes', assetIndexFile)
        local_assetIndexFile = os.path.join(self.assets_location, 'indexes', assetIndexFile)
        shutil.copy2(cached_assetIndexFile, local_assetIndexFile)

        for key, val in mc.get_assets_index()['objects'].items():
            h = val['hash']
            path = '{}/{}'.format(h[0:2], h)

            cached_assetfile = os.path.join(self.cache_location('minecraft'), 'assets', 'objects', path)
            local_assetfile = os.path.join(self.assets_location, 'objects', path)

            if not os.path.exists(os.path.dirname(local_assetfile)):
                os.makedirs(os.path.dirname(local_assetfile))

            shutil.copy2(cached_assetfile, local_assetfile)

    ##########################################################################

    def copy_forge(self, mc):
        forgedl = Forge(self.packlock.get_metadata('minecraft_version'),
                        self.packlock.get_metadata('forge_version'),
                        self.downloader.subdownloader('forge'))

        self.log.info('Installing forge and dependent libraries ...')
        forgedl.install_dependent_libraries(self.cache_location('library'), self.libs_location)

        installer_location = self.libs_location
        forgedl.install_installer_libraries(self.cache_location('library'), installer_location)

        forgedl.run_processors(installer_location, 'client')

##############################################################################
# THE END



#!/usr/bin/env python

import sys
if sys.version_info < (3, 5):
    sys.exit('Packmaker requires Python 3.5 or higher')

from setuptools import find_packages, setup

import os
try:
    packmaker_version = os.environ['PACKMAKER_VERSION']
except KeyError:
    import subprocess
    try:
        # try to build a PEP440 compatible version based on the output of
        # the git describe command
        def git_command(args):
            cmd = ['git', '-C', os.path.dirname(os.path.realpath(__file__))]
            return subprocess.check_output(cmd + args).decode().strip()
        v_full = git_command(['describe', '--tags', '--dirty=.dirty'])
        v_tag = git_command(['describe', '--tags', '--abbrev=0'])
        v_tail = v_full[len(v_tag):]
        packmaker_version = v_tag + v_tail.replace('-', '.dev', 1).replace('-', '+', 1)
    except subprocess.CalledProcessError:
        packmaker_version = 'unknown'

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
  name='packmaker',
  version=packmaker_version,
  description='minecraft modpack maker',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author='Mark Crewson',
  author_email='mark@crewson.net',
  url="http://minecraft.pages.routh.io/tools/packmaker",

  classifiers=[
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: POSIX :: Linux",
    "Topic :: Games/Entertainment",
  ],

  keywords='minecraft modded modpack',

  packages=find_packages(),

  install_requires=[
    'prettytable',
    'python-dateutil',
    'pyyaml',
    'requests',
    'requests_cache',
    'urllib3',
  ],

  python_requires='>=3',

  entry_points={
    'console_scripts': [
        'packmaker = packmaker.main:main'
    ]
  }
)



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import os
import shutil
import urllib.parse

from .base import BaseBuilder
from ..download import HttpDownloader, MavenDownloader
from ..forge import Forge
from ..framew.application import OperationError
from ..minecraft import Minecraft

##############################################################################


class RouthioBuilder (BaseBuilder):

    build_subloc = 'routhio'

    MAVEN_REPOS = ('https://libraries.minecraft.net/',
                   'https://repo1.maven.org/maven2/',
                   'https://files.minecraftforge.net/maven/'
                   )

    ##########################################################################

    def do_build(self):
        manifest = {'title': self.packlock.get_metadata('title'),
                    'name': self.packlock.get_metadata('name'),
                    'version': self.packlock.get_metadata('version'),
                    'gameVersion': self.packlock.get_metadata('minecraft_version'),
                    'minimumVersion': 2,
                    'objectsLocation': 'objects',
                    'librariesLocation': 'libraries',
                    'tasks': []
                    }

        copytasks = self.copy_local_files(self.build_location('objects'), self.packlock.yield_clientonly_files())
        modtasks, features = self.process_mods()

        if features:
            manifest['features'] = features

        manifest['tasks'] = modtasks
        manifest['tasks'].extend(copytasks)

        version_manifest = self.get_minecraft_version_manifest()
        if self.packlock.get_metadata('forge_version') is not None:
            self.process_forge(version_manifest)
        manifest['versionManifest'] = version_manifest

        routhio = self.packlock.get_extraopt('routhio')
        if routhio is not None and 'launch' in routhio:
            manifest['launch'] = routhio['launch']

        self.log.info('Writing the pack manifest file...')
        with open(os.path.join(self.build_location(), '{}.json'.format(self.packlock.get_metadata('name'))), 'w') as mf:
            mf.write(json.dumps(manifest, indent=2, sort_keys=True))

        if routhio is not None and 'html' in routhio:
            self.log.info('Copy launcher html files ...')
            self.copy_files(files_iterator=routhio['html'])

        pkgfilename = os.path.join(self.release_location(),
                                   '{}-{}-routhio.{}'.format(self.packlock.get_metadata('name'),
                                                             self.packlock.get_metadata('version'),
                                                             self.release_extension))
        self.log.info('Packaging the modpack: {} ...'.format(pkgfilename))
        self.release_pkg(pkgfilename)

    ##########################################################################

    def copy_local_files(self, dest, files_iterator):
        self.log.info('Copying local files ...')
        tasks = []
        for fileloc in files_iterator:
            for root, dirs, files in os.walk(fileloc['location']):
                for file in files:
                    full_filename = os.path.join(root, file)

                    sha1hash = self.calc_sha1(full_filename)
                    sha1dir = os.path.join(sha1hash[0:2], sha1hash[2:4])

                    dest_directory = os.path.join(dest, sha1dir)
                    dest_filename = os.path.join(dest_directory, sha1hash)

                    if not os.path.exists(dest_directory):
                        os.makedirs(dest_directory)

                    shutil.copy2(full_filename, dest_filename)

                    tasks.append({"type": 'file',
                                  "to": os.path.join(root[len(fileloc['location']):], file),
                                  "size": os.path.getsize(full_filename),
                                  "location": os.path.join(sha1dir, sha1hash),
                                  "hash": sha1hash
                                  })
        return tasks

    ##########################################################################

    def process_mods(self):
        tasks = []
        features = []
        self.download_mods(self.packlock.yield_clientonly_mods(include_optional=True))
        for mod in self.packlock.yield_clientonly_mods(include_optional=True):
            localfile = '{}/{}'.format(self.cache_location('addons'), mod.fileName)
            if not os.path.exists(localfile):
                raise OperationError('Missing local copy of mod even after downloading: {}'.format(localfile))

            sha1hash = self.calc_sha1(localfile)
            task = {'type': 'url',
                    'to': os.path.join('mods', mod.fileName),
                    'location': mod.downloadUrl,
                    'hash': sha1hash,
                    'size': os.path.getsize(localfile)
                    }

            moddef = self.packlock.get_mod(mod.slug)
            if moddef.optional:
                task['when'] = {'features': [mod.name], 'if': 'requireAny'}
                feature = {'name': mod.name,
                           'description': moddef.description,
                           'recommendation': moddef.recommendation,
                           'selected': moddef.selected
                           }
                features.append(feature)

            tasks.append(task)
        return tasks, features

    ##########################################################################

    def calc_sha1(self, filename):
        h = hashlib.sha1()
        with open(filename, 'rb') as hf:
            data = hf.read(65536)
            while data:
                h.update(data)
                data = hf.read(65536)
        return h.hexdigest()

    ##########################################################################

    def get_minecraft_version_manifest(self):
        mc = Minecraft(self.packlock.get_metadata('minecraft_version'))
        version_manifest = mc.get_version_manifest()

        # tidy up a bit
        version_manifest.pop('logging', None)
        version_manifest.pop('downloads', None)

        return version_manifest

    ##########################################################################

    def process_forge(self, version_manifest):
        downloader = HttpDownloader(self.cache_location('forge'))
        forgedl = Forge(self.packlock.get_metadata('minecraft_version'),
                        self.packlock.get_metadata('forge_version'),
                        downloader)
        self.log.info('Finding maven repositories for dependent libraries ...')
        mavenDownloader = MavenDownloader(self.cache_location('library'), self.MAVEN_REPOS)
        for library in forgedl.get_dependent_libraries():
            # cleanup the library entry a bit
            libdict = library.library
            libdict.pop('checksums', None)
            libdict.pop('serverreq', None)
            libdict.pop('clientreq', None)

            if libdict['name'].startswith('net.minecraftforge:forge:'):
                universal_url = forgedl.universal_download_url()
                universal_path = urllib.parse.urlparse(universal_url).path[1:]

                downloads = {'artifact': {'path': universal_path,
                                          'url': universal_url,
                                          }}
                libdict['downloads'] = downloads
                libdict.pop('url', None)

            else:
                # Find a maven repository hosting this library if necessary
                libdict['url'] = mavenDownloader.find_maven_repo(libdict['name'])

            version_manifest['libraries'].append(libdict)

        for override in ('mainClass', 'minecraftArguments'):
            o = forgedl.get_version_info().get(override, None)
            if o is not None:
                version_manifest[override] = o

##############################################################################
# THE END
